import numpy as np

from ... import Operation as op
from ...Operation import Operation
from ... import optools

class BgSubtractByTemperature(Operation):
    """
    Originally contributed by Amanda Fournier.  

    Find a background spectrum from a batch of background spectra,
    where the temperature of the background spectrum is as close as possible
    to the (input) temperature of the measured spectrum.
    Then subtract that background spectrum from the input spectrum.
    The measured and background spectra are expected to have the same domain.
    """
    
    def __init__(self):
        input_names = ['q_I_meas','T_meas','bg_batch_output','bg_I_uri','bg_T_uri']
        output_names = ['q_I_bgsub', 'T_bg', 'bg_factor']
        super(BgSubtractByTemperature, self).__init__(input_names, output_names)
        self.input_doc['q_I_meas'] = 'n-by-2 array of I(q) versus q'
        self.input_doc['T_meas'] = str('temperature as taken from the dict '
        + 'produced by the detector header file')
        self.input_doc['bg_batch_output'] = str('the output (list of dicts) '
        + 'of a batch of background spectra at different temperatures')
        self.input_doc['bg_I_uri'] = str('the uri for the items saved in bg_batch_output '
        + 'containing the background intensity spectrum, expected to be on same q domain as q_I_meas') 
        self.input_doc['bg_T_uri'] = str('the uri for the items saved in bg_batch_output '
        + 'containing the background spectrum temperatures')
        self.output_doc['q_I_bgsub'] = 'n by 2 array of q and background subracted intensity (I_meas-bg_factor*I_bg)'
        self.output_doc['T_bg'] = 'Temperature of the subtracted background spectrum'
        self.output_doc['bg_factor'] = str('correction factor applied to background '
        + 'before subtraction, to ensure positive intensity values')
        self.input_src['q_I_meas'] = op.wf_input
        self.input_src['T_meas'] = op.wf_input
        self.input_src['bg_batch_output'] = op.wf_input
        self.input_src['bg_I_uri'] = op.wf_input
        self.input_src['bg_T_uri'] = op.wf_input
        self.input_type['q_I_meas'] = op.ref_type
        self.input_type['T_meas'] = op.ref_type
        self.input_type['bg_batch_output'] = op.ref_type
        self.input_type['bg_T_uri'] = op.path_type
        self.input_type['bg_I_uri'] = op.path_type

    def run(self):
        q_I_meas = self.inputs['q_I_meas']
        q_I_bgsub = np.array(q_I_meas)
        T_meas = self.inputs['T_meas']
        bg_I_uri = self.inputs['bg_I_uri']
        bg_T_uri = self.inputs['bg_T_uri']
        bg_out = self.inputs['bg_batch_output']
        T_allbg = [optools.get_uri_from_dict(bg_T_uri,d) for d in bg_out]
        closest_T_idx = np.argmin(np.abs([T_bg - T_meas for T_bg in T_allbg]))
        T_bg = T_allbg[closest_T_idx]
        I_bg = optools.get_uri_from_dict(bg_I_uri,bg_out[closest_T_idx])
        #if not all(q_I[:,0] == q_I_bg[:,0]):
        #    msg = 'SPECTRUM AND BACKGROUND ON DIFFERENT q DOMAINS'
        #    raise ValueError(msg)
        bad_data = (q_I_meas[:,1] <= 0) | (I_bg <= 0) | np.isnan(q_I_meas[:,1]) | np.isnan(I_bg)
        I_floor = 0 
        #I_floor = 1E-6 * np.max(q_I_meas[:,1])
        bg_factor = np.min((q_I_meas[:,1][~bad_data]-I_floor) / I_bg[~bad_data])
        #print 'bg factor is {}'.format(bg_factor)
        q_I_bgsub[:,1] = q_I_meas[:,1] - bg_factor * I_bg

        #from matplotlib import pyplot as plt
        #plt.figure(1)
        #plt.semilogy(q_I_meas[:,0],q_I_meas[:,1])
        #plt.semilogy(q_I_meas[:,0],I_bg,'r')
        #plt.semilogy(q_I_bgsub[:,0],q_I_bgsub[:,1],'g')
        #plt.legend(['sample','bg','(sample - bg)'],loc=4)
        #print bg_factor
        #plt.show()

        self.outputs['q_I_bgsub'] = q_I_bgsub
        self.outputs['T_bg'] = T_bg 
        self.outputs['bg_factor'] = bg_factor





from collections import OrderedDict

from ... import Operation as op
from ...Operation import Operation

class ReadPONI(Operation):
    """
    Read in a dict of PyFAI PONI parameters.
    Input path to a .poni file representing a calibrated measurement geometry.
    """
    
    def __init__(self):
        input_names = ['poni_file']
        output_names = ['poni_dict']
        super(ReadPONI,self).__init__(input_names,output_names)
        self.input_doc['poni_file'] = '.poni file describing a calibration result'
        self.input_src['poni_file'] = op.fs_input
        self.input_type['poni_file'] = op.path_type
        self.output_doc['poni_dict'] = 'Dict of pyFAI calibration parameters, as found in a .poni file'

    def run(self):
        fpath = self.inputs['poni_file']
        pdict = OrderedDict()
        for line in open(fpath,'r'):
            kv = line.strip().split('=')
            if kv[0] == 'dist':
                pdict['dist'] = float(kv[1])         
            if kv[0] == 'poni1':
                pdict['poni1'] = float(kv[1])       
            if kv[0] == 'poni2':
                pdict['poni2'] = float(kv[1])       
            if kv[0] == 'rot1':    
                pdict['rot1'] = float(kv[1])       
            if kv[0] == 'rot2':    
                pdict['rot2'] = float(kv[1])       
            if kv[0] == 'rot3':    
                pdict['rot3'] = float(kv[1])       
            if kv[0] == 'pixel1':    
                pdict['pixel1'] = float(kv[1])       
            if kv[0] == 'pixel2':    
                pdict['pixel2'] = float(kv[1])       
            if kv[0] == 'wavelength':    
                pdict['wavelength'] = float(kv[1])       
            if kv[0] == 'fpolz':    
                pdict['fpolz'] = float(kv[1])       
            if kv[0] == 'detector':    
                pdict['detector'] = kv[1]
            if kv[0] == 'splineFile':    
                pdict['splineFile'] = kv[1]
        self.outputs['poni_dict'] = pdict 




import os
from functools import partial
from collections import OrderedDict

from PySide import QtGui, QtCore, QtUiTools
import yaml

from ..core.operations.Operation import Operation
from . import uitools
from . import widgets 
from ..core import pawstools
from .WfUiManager import WfUiManager
from .OpUiManager import OpUiManager
from .PluginUiManager import PluginUiManager
from ..core.models.ListModel import ListModel
from ..core.plugins.WfManagerPlugin import WfManagerPlugin
from ..qt.QWfManager import QWfManager
from ..qt.QOpManager import QOpManager
from ..qt.QPluginManager import QPluginManager

class UiManager(QtCore.QObject):
    """
    Stores a reference to a QMainWindow,
    performs operations on it
    """

    def __init__(self,paws_api,app):
        super(UiManager,self).__init__()
        ui_file = QtCore.QFile(pawstools.sourcedir+"/ui/qtui/basic.ui")
        ui_file.open(QtCore.QFile.ReadOnly)
        self.ui = QtUiTools.QUiLoader().load(ui_file)
        ui_file.close()
        self.ui.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        paws_api._op_manager.logmethod = self.msg_board_log
        paws_api._wf_manager.logmethod = self.msg_board_log
        paws_api._plugin_manager.logmethod = self.msg_board_log
        self.qplugman = QPluginManager(paws_api._plugin_manager)
        self.qopman = QOpManager(paws_api._op_manager)
        self.qwfman = QWfManager(paws_api._wf_manager,app)
        self.paw = paws_api
        self.make_title()
        self.build()

    @QtCore.Slot(str)
    def update_wfman_plugin(self,wfname):
        self.qplugman.update_plugin('wf_manager')

    #requestStopWorkflow = QtCore.Signal(str)
    #requestRunWorkflow = QtCore.Signal(str)

    def msg_board_log(self,msg):
        """Print timestamped message to msg board"""
        if (self.ui.message_board.verticalScrollBar().value() == 
        self.ui.message_board.verticalScrollBar().maximum()):
            advance_scrollbar = True
        else:
            advance_scrollbar = False
        self.ui.message_board.appendPlainText(
        '- ' + pawstools.timestr() + ': ' + msg)#+ '\n') 
        if advance_scrollbar:
            self.ui.message_board.verticalScrollBar().setValue(
            self.ui.message_board.verticalScrollBar().maximum())

    def make_title(self):
        """Display the paws logo in the image viewer"""
        img_file = os.path.join(pawstools.sourcedir, "ui/graphics/paws_icon_white.png")
        pixmap = QtGui.QPixmap(img_file)
        pixmap_item = QtGui.QGraphicsPixmapItem(pixmap)
        scene = QtGui.QGraphicsScene()
        scene.addItem(pixmap_item)
        qwhite = QtGui.QColor(255,255,255,255)
        textitem = scene.addText("v{}".format(pawstools.version))
        textitem.setPos(100,35)
        textitem.setDefaultTextColor(qwhite)
        logo_view = QtGui.QGraphicsView()
        logo_view.setScene(scene)
        self.ui.viewer_layout.addWidget(logo_view,0,0,1,1)
        self.ui.setWindowTitle("paws v{}".format(pawstools.version))
        self.ui.setWindowIcon(pixmap)

    def build(self):
        """
        Set up QObjects and model views 
        for communicating with paws objects 
        """
        self.qwfman.wf_updated.connect( partial(self.update_wfman_plugin) )
        self.qwfman.wfdone.connect(self.update_run_wf_button)
        #self.requestRunWorkflow.connect(self.qwfman.runWorkflow)
        #self.requestStopWorkflow.connect(self.qwfman.stopWorkflow)
        lm = ListModel(self.qwfman.qworkflows.keys())
        self.ui.wf_selector.setModel(lm)
        self.ui.wf_selector.currentIndexChanged.connect( partial(self.set_wf) )
        self.ui.add_wf_button.clicked.connect( partial(self.add_wf,'new_workflow') )
        self.ui.plugin_tree.setModel(self.qplugman)
        self.ui.op_tree.setModel(self.qopman)
        #self.ui.op_tree.clicked.connect( partial(uitools.toggle_expand,self.ui.op_tree) ) 
        self.ui.op_tree.setRootIndex(self.qopman.root_index())
        #self.ui.wf_tree.clicked.connect( partial(uitools.toggle_expand,self.ui.wf_tree) )
        self.ui.op_tree.clicked.connect( self.display_op_item )
        self.ui.wf_tree.clicked.connect( self.display_wf_item )
        self.ui.wf_tree.doubleClicked.connect( partial(self.edit_wf) )
        self.ui.plugin_tree.clicked.connect( self.display_plugin_item )
        self.ui.plugin_tree.setRootIndex(self.qplugman.root_index())
        self.ui.run_wf_button.setText("&Run")
        self.ui.run_wf_button.clicked.connect(self.toggle_run_wf)
        self.ui.edit_ops_button.setText("Edit operations")
        self.ui.edit_ops_button.clicked.connect(self.edit_ops)
        self.ui.add_wf_button.setText("&New workflow")
        self.ui.edit_wf_button.setText("&Edit workflow")
        self.ui.edit_wf_button.clicked.connect(self.edit_wf)
        self.ui.save_wf_button.setText("&Save workflow")
        self.ui.save_wf_button.clicked.connect(self.save_wf)
        self.ui.load_wf_button.setText("&Load workflow")
        self.ui.load_wf_button.clicked.connect(self.load_wf)
        self.ui.save_plugins_button.setText("Save plugins")
        self.ui.save_plugins_button.clicked.connect(self.save_plugins)
        self.ui.load_plugins_button.setText("Load plugins")
        self.ui.load_plugins_button.clicked.connect(self.load_plugins)
        self.ui.edit_plugins_button.setText("Edit plugins")
        self.ui.edit_plugins_button.clicked.connect(self.start_plugins_ui)
        self.ui.message_board.setReadOnly(True)
        self.ui.message_board.insertPlainText('--- MESSAGE BOARD ---') 
        self.msg_board_log('paws is ready') 
        #self.ui.op_tree.hideColumn(1)
        #self.ui.op_tree.hideColumn(2)
        self.ui.op_tree.setColumnWidth(0,200)
        self.ui.plugin_tree.hideColumn(1)
        self.ui.plugin_tree.hideColumn(2)
        self.ui.hsplitter.setStretchFactor(1,2)    
        self.ui.vsplitter.setStretchFactor(0,1)    

    def set_wf(self,wf_selector_idx):
        wfname = self.ui.wf_selector.model().list_data()[wf_selector_idx]
        self.paw.select_wf(wfname)
        self.ui.wf_tree.setModel(self.qwfman.qworkflows[wfname])
        self.ui.wf_tree.setRootIndex(self.qwfman.qworkflows[wfname].root_index())
        self.update_run_wf_button()

    def add_wf(self,wfname):
        #if wfname in self.wfman.workflows.keys():
        #    wfname = self.wfman.auto_name(wfname)
        if wfname in self.qwfman.qworkflows.keys():
            raise KeyError('[{}] Name {} already assigned to a Workflow. '
            .format(__name__,wfname) + 'Loaded workflows: {}'
            .format(self.qwfman.qworkflows.keys()))
        self.qwfman.add_wf(wfname)
        self.ui.wf_selector.model().append_item(wfname)
        # if this is the first workflow loaded, need to hide the treeview columns.
        if self.qwfman.n_wf() == 1:
            self.ui.wf_tree.hideColumn(1)
            self.ui.wf_tree.hideColumn(2)
        self.ui.wf_selector.setCurrentIndex(self.qwfman.n_wf()-1)

    def display_op_item(self,idx):
        """
        Display selected item from the op tree in viewer layout 
        """
        if idx.isValid(): 
            # If the user is clicking a check box (idx.column()!=0),
            # leave the central display unchanged... ?
            #if idx.column() == 0:
                itm_data = self.qopman.get_data_from_index(idx)
                itm_uri = self.qopman.get_uri_of_index(idx)
                w = QtGui.QTextEdit()
                if itm_data is None:
                    # this should be the condition for a not-enabled Operation
                    t = str('selected Operation: {}. \n'.format(itm_uri)
                    + 'This Operation is not currently enabled. '
                    + 'Click the "enable" box to enable it. ')
                elif isinstance(itm_data,dict):
                    # this should be the condition for a category
                    t = str('selected category: {}. \n\n'.format(itm_uri)
                    + '{}: '.format(itm_uri)) 
                    t = t + self.qopman.opman.print_cat(itm_uri) 
                else:
                    # the only remaining case is an enabled Operation
                    t = itm_data().description()
                w.setText(t)
                self.display_widget(w)

    def display_plugin_item(self,idx):
        """
        Display selected item from the plugin tree in viewer layout 
        """
        if idx.isValid(): 
            itm_data = self.qplugman.get_data_from_index(idx)
            widg = widgets.make_widget(itm_data)
            self.display_widget(widg)

    def display_wf_item(self,idx):
        """
        Display selected item from the workflow tree in viewer layout 
        """
        if idx.isValid(): 
            itm_data = self.current_wf().get_data_from_index(idx)
            widg = widgets.make_widget(itm_data)
            self.display_widget(widg)

    def display_widget(self,widg):
        # Loop through the viewer layout, last to first, clear the frame
        n_widgets = self.ui.viewer_layout.count()
        for i in range(n_widgets-1,-1,-1):
            # QLayout.takeAt returns a LayoutItem
            itm = self.ui.viewer_layout.takeAt(i)
            # get the QWidget of that LayoutItem and close it 
            itm.widget().close()
            #if itm.widget() is not None:
            #    itm.widget().deleteLater()
        self.ui.viewer_layout.addWidget(widg,0,0,1,1) 

    def current_wf(self):
        wfname = self.current_wfname()
        if wfname:
            return self.qwfman.qworkflows[wfname]

    def current_wfname(self):
        idx = self.ui.wf_selector.currentIndex()
        if idx == -1:
            return None
        else:
            return self.ui.wf_selector.model().list_data()[idx]

    def toggle_run_wf(self,wfname=None):
        if wfname:
            qwf = self.qwfman.qworkflows[wfname]
        else:
            wfname = self.current_wfname()
            qwf = self.qwfman.qworkflows[wfname]
        if self.qwfman.wf_running[wfname]: 
            self.qwfman.stop_wf(wfname)
            self.ui.run_wf_button.setText("&Run")
            #self.requestStopWorkflow.emit(wfname)
        else:
            self.ui.run_wf_button.setText("S&top")
            self.qwfman.run_wf(wfname)
        #self.update_run_wf_button()

    def update_run_wf_button(self):
        wfname = self.current_wfname() 
        if self.qwfman.wf_running[wfname]:
            self.ui.run_wf_button.setText("S&top")
        else:
            self.ui.run_wf_button.setText("&Run")

    def edit_wf(self,itm_idx=QtCore.QModelIndex()):
        """
        Interact with user to edit the workflow.
        Pass in QModelIndex to open the editor 
        with the item at that index loaded.
        """
        wf = self.current_wf()
        if wf is None:
            self.add_wf('new_workflow')
            wf = self.current_wf()
        if itm_idx.isValid():
            # valid index in workflow tree: percolate up to root ancestor
            while itm_idx.parent().isValid():
                itm_idx = itm_idx.parent()
        if itm_idx.isValid():
            uiman = self.start_wf_editor(wf,itm_idx)
        else:
            uiman = self.start_wf_editor(wf)
        uiman.ui.show()

    def start_wf_editor(self,qwf=None,idx=QtCore.QModelIndex()):
        """
        Create a WfUiManager (QMainWindow), return it 
        """
        uiman = WfUiManager(self.qwfman,self.qopman,self.qplugman)
        uiman.ui.wf_selector.setCurrentIndex(self.ui.wf_selector.currentIndex())
        #uiman.set_wf()
        if qwf and idx.isValid():
            uiman.get_op(qwf,idx)
        uiman.ui.setParent(self.ui,QtCore.Qt.Window)
        return uiman

    def edit_ops(self,itm_idx=None):
        """
        interact with user to enable existing Operations
        and edit or develop new Operations 
        """
        uiman = OpUiManager(self.qopman)
        uiman.ui.setParent(self.ui,QtCore.Qt.Window)
        uiman.ui.show()

    def start_plugins_ui(self):
        uiman = PluginUiManager(self.qplugman)
        uiman.ui.setParent(self.ui,QtCore.Qt.Window)
        uiman.ui.show()

    def save_plugins(self):
        """
        Start a modal window dialog to choose a .wfl to save the current set of plugins  
        """
        save_ui = uitools.start_save_ui(self.ui)
        save_ui.setWindowTitle('plugins saver')
        save_ui.tree_box.setTitle('Select or enter a .wfl file to save current plugins')
        save_ui.save_button.clicked.connect( partial(self.finish_save_plugins,save_ui) )
        save_ui.show()
        save_ui.activateWindow()

    def load_plugins(self):
        """
        Start a modal window dialog to choose a .wfl to load a set of plugins 
        """
        load_ui = uitools.start_load_ui(self.ui)
        load_ui.setWindowTitle('plugins loader')
        load_ui.tree_box.setTitle('Select a .wfl file to load plugins')
        load_ui.load_button.clicked.connect( partial(self.finish_load_plugins,load_ui) )
        load_ui.show()
        load_ui.activateWindow()
      
    def save_wf(self):
        """
        Start a modal window dialog to choose a .wfl to save the current workflow  
        """
        save_ui = uitools.start_save_ui(self.ui)
        save_ui.setWindowTitle('workflow saver')
        wf_idx = self.ui.wf_selector.currentIndex()
        wfname = self.ui.wf_selector.model().list_data()[wf_idx]
        save_ui.tree_box.setTitle('Select or enter a .wfl file to save workflow setup for {}'.format(wfname))
        save_ui.save_button.clicked.connect( partial(self.finish_save_wf,save_ui) )
        save_ui.show()
        save_ui.activateWindow()

    def load_wf(self):
        """
        Start a modal window dialog to choose a .wfl to load a workflow
        """
        load_ui = uitools.start_load_ui(self.ui)
        load_ui.setWindowTitle('workflow loader')
        load_ui.tree_box.setTitle('Select a .wfl file to load a workflow')
        load_ui.load_button.clicked.connect( partial(self.finish_load_wf,load_ui) )
        load_ui.show()
        load_ui.activateWindow()
      
    def finish_save_plugins(self,ui):
        fname = ui.filename.text()
        if not os.path.splitext(fname)[1] == '.wfl':
            fname = fname + '.wfl'
        self.msg_board_log( 'dumping current set of plugins to {}'.format(fname) )
        d = {} 
        pgin_dict = OrderedDict() 
        for pgin_name in self.qplugman.plugman.list_plugins():
            pgin = self.qplugman.plugman.get_data_from_uri(pgin_name)
            if not isinstance(pgin,WfManagerPlugin):
                pgin_dict[pgin_name] = self.qplugman.plugman.plugin_setup_dict(pgin)
        d['PLUGINS'] = pgin_dict
        pawstools.update_file(fname,d)
        ui.close()

    def finish_load_plugins(self,ui):
        fname = ui.tree.model().filePath(ui.tree.currentIndex())
        f = open(fname,'r')
        d = yaml.load(f)
        f.close()
        fname_nopath = os.path.split(fname)[1]
        fname_noext = os.path.splitext(fname_nopath)[0]
        if 'PLUGINS' in d.keys():
            self.qplugman.load_from_dict(d['PLUGINS'])
        ui.close()

    def finish_save_wf(self,ui):
        # TODO: Should a saved workflow change its name?
        # Would have to update ui.wf_selector
        # and the WfManager or QWfManager 
        fname = ui.filename.text()
        self.paw.save_workflow(fname)
        ui.close()

    def finish_load_wf(self,ui):
        fname = ui.tree.model().filePath(ui.tree.currentIndex())
        f = open(fname,'r')
        d = yaml.load(f)
        f.close()
        fname_nopath = os.path.split(fname)[1]
        fname_noext = os.path.splitext(fname_nopath)[0]
        if 'WORKFLOW' in d.keys():
            wfname = fname_noext
            self.qwfman.load_from_dict(wfname,self.qopman.opman,d['WORKFLOW'])
            #if wfname in self.qwfman.qworkflows.keys():
            #    wfname = self.qwfman.auto_name(wfname)
            if not wfname in self.ui.wf_selector.model().list_data():
                self.ui.wf_selector.model().append_item(wfname)
            new_wf_idx = self.ui.wf_selector.model().list_data().index(wfname) 
            if self.qwfman.n_wf() == 1:
                self.ui.wf_tree.hideColumn(1)
                self.ui.wf_tree.hideColumn(2)
        self.ui.wf_selector.setCurrentIndex(new_wf_idx)
        ui.close()




import fabio

from ... import Operation as op
from ...Operation import Operation

class FabIOOpen(Operation):
    """
    Takes a filesystem path and calls fabIO to load it. 
    """

    def __init__(self):
        input_names = ['path']
        output_names = ['image_data']
        super(FabIOOpen,self).__init__(input_names,output_names) 
        self.input_doc['path'] = 'string representing the path to a .tif image'
        self.output_doc['image_data'] = '2D array representing pixel values taken from the input file'
        self.input_src['path'] = op.fs_input
        self.input_type['path'] = op.path_type
        
    def run(self):
        """
        Call on fabIO to extract image data
        """
        img_url = self.inputs['path']
        self.outputs['image_data'] = fabio.open(img_url).data




from collections import OrderedDict
import copy
from functools import partial

from ..models.TreeModel import TreeModel
from ..operations import Operation as op
from ..operations.Operation import Operation, Batch, Realtime

class Workflow(TreeModel):
    """
    Tree structure for a Workflow built from paws Operations.
    """

    def __init__(self):
        super(Workflow,self).__init__()
        #self.wfman = wfman

    def __getitem__(self,key):
        optags = self.keys()
        if key in optags:
            return self.get_data_from_uri(key) 
        else:
            raise KeyError('[{}] {}.__getitem__ only recognizes keys {}'
            .format(__name__,type(self).__name__,optags))
    def __setitem__(self,key,data):
        optags = self.keys() 
        # TODO: ensure that data is an Operation?
        if key in optags:
            self.set_item(key,data)
        else:
            raise KeyError('[{}] {}.__setitem__ only recognizes keys {}'
            .format(__name__,type(self).__name__,optags))
    def keys(self):
        return self.list_op_tags() 

    def build_tree(self,x):
        """
        Reimplemented TreeModel.build_tree() 
        so that TreeItems are built from Operations.
        """
        if isinstance(x,Operation):
            d = OrderedDict()
            d[op.inputs_tag] = self.build_tree(x.inputs)
            d[op.outputs_tag] = self.build_tree(x.outputs)
            return d
        else:
            return super(Workflow,self).build_tree(x) 

    def op_dict(self):
        optags = self.list_op_tags() 
        op_dict = OrderedDict(zip(optags,[self.get_data_from_uri(nm) for nm in optags]))
        return op_dict

    def list_op_tags(self):
        return self.root_tags()

    def n_ops(self):
        return self.n_children()

    def set_op_input_at_uri(self,uri,val):
        """
        Set an op input at uri to provided value val.
        The uri must be a valid uri in the TreeModel,
        of the form opname.inputs.inpname.
        """
        path = uri.split('.')
        opname = path[0]
        if not path[1] == op.inputs_tag:
            msg = '[{}] uri {} does not point to an input'.format(__name__,uri)
            raise ValueError(msg)
        inpname = path[2]
        uri = opname+'.'+op.inputs_tag+'.'+inpname
        op = self.get_data_from_uri(opname)
        op.input_locator[inpname].data = val
        self.set_item(uri,val)





import abc
import re
from collections import OrderedDict

# constants used for enumerative types etc
# tags for inputs and outputs TreeItems    
inputs_tag = 'inputs'
outputs_tag = 'outputs'

# Declarations of valid sources and types for workflow and plugin inputs
##### TODO: the following but gracefully
no_input = 0
text_input = 1
fs_input = 2
wf_input = 3
plugin_input = 4
batch_input = 5 
valid_sources = [no_input,text_input,fs_input,wf_input,plugin_input,batch_input]
input_sources = ['none','text','filesystem','workflow','plugins','batch'] 

none_type = 0
str_type = 1
int_type = 2
float_type = 3
bool_type = 4
ref_type = 5
path_type = 6
auto_type = 7
valid_types = [none_type,str_type,int_type,float_type,bool_type,ref_type,path_type,auto_type]
input_types = ['none','string','integer','float','boolean','reference','path','auto']

invalid_types = {}                
invalid_types[no_input] = [str_type,int_type,float_type,bool_type,ref_type,path_type,auto_type]
invalid_types[text_input] = [ref_type,path_type,auto_type]
invalid_types[fs_input] = [str_type,int_type,float_type,bool_type,ref_type,auto_type]
invalid_types[wf_input] = [str_type,int_type,float_type,bool_type,auto_type]
invalid_types[plugin_input] = [str_type,int_type,float_type,bool_type,auto_type]
invalid_types[batch_input] = [str_type,int_type,float_type,bool_type,ref_type,path_type]

class InputLocator(object):
    """
    Objects of this class are used as containers for inputs to an Operation.
    They contain the information needed to find the relevant input data.
    After the data is loaded, it should be stored in InputLocator.data.
    """
    def __init__(self,src=no_input,tp=none_type,val=None):
        self.src = src
        self.tp = tp
        self.val = val 
        self.data = None 

def parameter_doc(name,value,doc):
    if isinstance(value, InputLocator):
        src_str = input_sources[value.src]
        tp_str = input_types[value.tp]
        v_str = str(value.val)
        return "- name: {} \n- source: {} \n- type: {} \n- value: {} \n- doc: {}".format(name,src_str,tp_str,v_str,doc) 
    else:
        tp_str = type(value).__name__
        return "- name: {} \n- type: {} \n- doc: {}".format(name,tp_str,doc) 

class Operation(object):
    """
    Class template for implementing paws operations.
    """

    def __init__(self,input_names,output_names):
        """
        The input_names and output_names (lists of strings)
        are used to specify names for the parameters 
        that will be used to perform the operation.
        These lists are used as keys to build dicts
        Operation.inputs and Operation.outputs.
        """
        self.inputs = OrderedDict()
        self.input_locator = OrderedDict() 
        self.outputs = OrderedDict() 
        self.input_doc = OrderedDict() 
        self.input_src = OrderedDict() 
        self.input_type = OrderedDict() 
        self.output_doc = OrderedDict() 
        # For each of the i/o names, assign to None 
        for name in input_names: 
            self.input_src[name] = no_input
            self.input_type[name] = none_type
            self.input_locator[name] = None 
            self.inputs[name] = None
            self.input_doc[name] = None
        for name in output_names: 
            self.outputs[name] = None
            self.output_doc[name] = None
        # Set flags so that Batch and Realtime Operations
        # can be identified without having to import them.
        # Override these flags in Batch and Realtime.
        self._batch_flag = False
        self._realtime_flag = False

    def __getitem__(self,key):
        if key == inputs_tag:
            return self.inputs
        elif key == outputs_tag:
            return self.outputs
        else:
            raise KeyError('[{}] Operation only recognizes keys {}'
            .format(__name__,self.keys()))
    def __setitem__(self,key,data):
        if key == inputs_tag:
            self.inputs = data
        elif key == outputs_tag:
            self.outputs = data
        else:
            raise KeyError('[{}] Operation only recognizes keys {}'
            .format(__name__,self.keys()))
    def keys(self):
        return [inputs_tag,outputs_tag]

    def load_defaults(self):
        for name in self.inputs.keys():
            src = no_input
            tp = none_type
            val = None
            if not self.input_src[name] == no_input:
                src = self.input_src[name]
                #if (self.input_type[name] == none_type
                #and src in [wf_input,fs_input,batch_input,plugin_input]):
                #    self.input_type[name] = auto_type
                if (not self.input_type[name] == none_type
                and not self.input_type[name] in invalid_types[src]):
                    tp = self.input_type[name]
                    if self.inputs[name] is not None:
                        if isinstance(self.inputs[name],list):
                            val = [str(v) for v in self.inputs[name]]
                        else:
                            val = str(self.inputs[name])
            self.input_locator[name] = InputLocator(src,tp,val)
            # defaults are now packaged in InputLocators, so can be dereferenced from self.inputs. 
            self.inputs[name] = None

    def run(self):
        """
        Operation.run() should use all of the items in Operation.inputs
        and set values for all of the items in Operation.outputs.
        """
        pass

    def description(self):
        """
        self.description() returns a string 
        documenting the input and output structure 
        and usage instructions for the Operation
        """
        return str(
        "Operation description: "
        + self.doc_as_string()
        + "\n\n--- Inputs ---"
        + self.input_description() 
        + "\n\n--- Outputs ---"
        + self.output_description())

    def doc_as_string(self):
        if self.__doc__:
            return re.sub("\s\s+"," ",self.__doc__.replace('\n','')) 
        else:
            return "none"

    def input_description(self):
        a = ""
        inp_indx = 0
        for name in self.inputs.keys(): 
            if self.input_locator[name]:
                display_val = self.input_locator[name]
            else:
                display_val = self.inputs[name] 
            a = a + str("\n\nInput {}:\n".format(inp_indx) 
            + parameter_doc(name,display_val,self.input_doc[name]))
            inp_indx += 1
        return a

    def output_description(self):
        a = ""
        out_indx = 0
        for name,val in self.outputs.items(): 
            a = a + str("\n\nOutput {}:\n".format(out_indx) 
            + parameter_doc(name,val,self.output_doc[name]))
            out_indx += 1
        return a
                
class Batch(Operation):
    __metaclass__ = abc.ABCMeta
    """
    Class template for implementing batch execution operations.
    """
    def __init__(self,input_names,output_names):
        super(Batch,self).__init__(input_names,output_names)
        # Override Operation._realtime_flag
        # so Batch can be identified without importing
        self._batch_flag = True


    @abc.abstractmethod
    def output_list(self):
        """
        Produce a list of OrderedDicts representing the outputs for each batch input.
        Each OrderedDict should be populated with [output_uri:output_value] pairs.
        """
        pass

    @abc.abstractmethod
    def input_list(self):
        """
        Produce a list of OrderedDicts representing each set of inputs for the Batch to run.
        Each OrderedDict should be populated with [input_uri:input_value] pairs.
        """
        pass

    @abc.abstractmethod
    def input_routes(self):
        """
        Produce a list of the input routes used by the Batch,
        in the same order as each of the OrderedDicts 
        provided by Batch.input_list().
        """
        pass

    @abc.abstractmethod
    def batch_outputs_tag(self):
        """
        Return the output name (one of the self.outputs.keys()) 
        that indicates where the batch outputs should be stored. 
        """
        pass
    
    @abc.abstractmethod
    def saved_items(self):
        """
        Return a list of items (as workflow uris) 
        to be saved after each execution.
        """
        pass 

    @abc.abstractmethod
    def set_batch_ops(self,wf=None):
        """
        Set enough information in this Operation's inputs
        so that self.batch_ops() returns the correct
        list of operations to be run under the batch.
        Takes a Workflow as optional second argument,
        so that it can be used to call optools.locate_input()
        """
        pass

    @abc.abstractmethod
    def set_input_routes(self,wf=None):
        """
        Set enough information in this Operation's inputs
        so that self.input_routes() returns the correct
        list of workflow uris where the batch will set its inputs 
        Takes a Workflow as optional second argument,
        so that it can be used to call optools.locate_input()
        """
        pass


class Realtime(Operation):
    __metaclass__ = abc.ABCMeta
    """
    Class template for implementing realtime execution as an Operation.
    """
    def __init__(self,input_names,output_names):
        super(Realtime,self).__init__(input_names,output_names)
        # Override Operation._realtime_flag
        # so realtime Operation can be identified without importing
        self._batch_flag = False
        self._realtime_flag = True

    @abc.abstractmethod
    def input_iter(self):
        """
        Produce an iterator over OrderedDicts representing each set of inputs to run.
        Each dict should be populated with [input_uri:input_value] pairs.
        When there is no new set of inputs to run, input_iter().next() should return None.
        """
        pass

    @abc.abstractmethod
    def input_routes(self):
        """
        Produce a list of the input routes used by the Realtime,
        in the same order as each of the OrderedDicts 
        provided by Realtime.input_iter().
        """
        pass

    @abc.abstractmethod
    def output_list(self):
        """
        Produce a list of OrderedDicts representing the outputs for each Realtime input.
        Each OrderedDict should be populated with [output_uri:output_value] pairs.
        """
        pass

    @abc.abstractmethod
    def batch_outputs_tag(self):
        """
        Return the output name (one of the self.outputs.keys()) 
        that indicates where the outputs should be stored. 
        """
        pass

    @abc.abstractmethod
    def saved_items(self):
        """
        Return a list of items (as workflow uris) 
        to be saved after each execution.
        """
        pass 

    def delay(self):
        """
        Return the number of MILLIseconds to pause between iterations.
        Overload this method to change the pause time- default is 1 second.
        """
        return 1000





import numpy as np
import pypif.obj as pifobj

from ... import Operation as op
from ...Operation import Operation

class PifTernary(Operation):
    """
    Package results from ternary wafer synthesis into a pypif.obj.ChemicalSystem object.
    """

    def __init__(self):
        input_names = ['name','q_I','q_texture','I_features','pk_idx','date_time','t_utc']
        output_names = ['pif']
        super(PifTernary,self).__init__(input_names,output_names)
        self.input_doc['name'] = 'user input string used as pif record uid'
        self.input_doc['q_I'] = 'array of q and I(q) values for xrd spectrum'
        self.input_doc['q_texture'] = 'array of q and texture(q) for xrd spectrum'
        self.input_doc['I_features'] = 'output of IntensityFeatures operation, providing Imax, Iave, Imax_Iave_ratio'
        self.input_doc['pk_idx'] = 'indices in q,I where peaks can be found'
        self.input_doc['date_time'] = 'string date/time from measurement header file for pif record tags'
        self.input_doc['t_utc'] = 'time in seconds utc'
        self.input_src['name'] = op.text_input
        self.input_src['q_I'] = op.wf_input
        self.input_src['q_texture'] = op.wf_input
        self.input_src['I_features'] = op.wf_input
        self.input_src['pk_idx'] = op.wf_input
        self.input_src['date_time'] = op.wf_input
        self.input_src['t_utc'] = op.wf_input
        self.input_type['name'] = op.str_type
        self.input_type['q_I'] = op.ref_type
        self.input_type['q_texture'] = op.ref_type
        self.input_type['I_features'] = op.ref_type
        self.input_type['pk_idx'] = op.ref_type
        self.input_type['date_time'] = op.ref_type
        self.input_type['t_utc'] = op.ref_type
        self.output_doc['pif'] = 'pif object containing the relevant data for this experiment'

    def run(self):
        uid_pre = self.inputs['name']
        q_I = self.inputs['q_I']
        q_tex = self.inputs['q_texture']
        I_feats = self.inputs['I_features']
        pk_idx = self.inputs['pk_idx']
        t_str = self.inputs['date_time']
        t_utc = self.inputs['t_utc']
        uid_full = uid_pre+'_'+str(int(t_utc))
        q = q_I[:,0]
        I_of_q = q_I[:,1]
        q_pk = np.array([q[idx] for idx in pk_idx])
        I_pk = np.array([I_of_q[idx] for idx in pk_idx])
        main_sys = pifobj.ChemicalSystem()
        main_sys.uid = uid_full
        main_sys.properties = self.xrd_to_pif_properties(q,I_of_q)
        main_sys.properties += self.texture_to_pif_properties(q_tex[:,0],q_tex[:,1])
        main_sys.properties += self.pks_to_pif_properties(q_pk,I_pk)
        main_sys.properties += self.intensity_features_to_pif_properties(I_feats)
        main_sys.tags = ['sample id: '+uid_pre,'date: '+t_str,'utc: '+str(int(t_utc))]
        self.outputs['pif'] = main_sys

    def intensity_features_to_pif_properties(self,I_feats):
        pImax = pifobj.Property()
        pImax.name = 'maximum XRD intensity'
        pImax.scalars = [ pifobj.Scalar(I_feats['Imax']) ]
        pImax.units = 'counts'
        pIave = pifobj.Property()
        pIave.name = 'average XRD intensity'
        pIave.scalars = [ pifobj.Scalar(I_feats['Iave']) ]
        pIave.units = 'counts'
        pratio = pifobj.Property()
        pratio.name = 'maximum-over-average XRD intensity'
        pratio.scalars = [ pifobj.Scalar(I_feats['Imax_Iave_ratio']) ]
        pratio.units = 'unitless'
        return [pImax,pIave,pratio] 

    def pks_to_pif_properties(self,q_pk,I_pk):
        props = []
        for i in range(len(q_pk)):
            p = pifobj.Property()
            p.name = 'XRD peak'
            p.scalars = [ pifobj.Scalar(I_pk[i]) ]
            p.conditions = [] 
            p.conditions.append( pifobj.Value('scattering vector',[pifobj.Scalar(q_pk[i])],None,None,'Angstrom^-1') )
            p.units = 'counts'
            props.append(p)
        return props

    def texture_to_pif_properties(self,q,tex_q):
        props = []
        for i in range(len(q)):
            p = pifobj.Property()
            p.name = 'texture'
            p.scalars = [ pifobj.Scalar(tex_q[i]) ]
            p.conditions = [] 
            p.conditions.append( pifobj.Value('scattering vector',[pifobj.Scalar(q[i])],None,None,'Angstrom^-1') )
            p.units = 'counts'
            props.append(p)
        return props

    def xrd_to_pif_properties(self,q,I_q):
        props = []
        for i in range(len(q)):
            p = pifobj.Property()
            p.name = 'XRD intensity'
            p.scalars = [ pifobj.Scalar(I_q[i]) ]
            p.conditions = [] 
            p.conditions.append( pifobj.Value('scattering vector',[pifobj.Scalar(q[i])],None,None,'Angstrom^-1') )
            p.units = 'counts'
            props.append(p)
        return props
        
#    def make_piftemperature(self,t):
#        v = pifobj.Value()
#        v.name = 'temperature'
#        tscl = pifobj.Scalar()
#        tscl.value = str(t)
#        v.scalars = [tscl]
#        v.units = 'degrees Celsius'
#        return v
#
#    def make_pifvector(self,v):
#        pifv = []
#        for n in v:
#            s = pifobj.Scalar()
#            s.value = str(n)
#            pifv.append(s)
#        return pifv
#
#    def pifscalar(self,scl,errlo,errhi):
#        s = pifobj.Scalar()
#        s.value = str(scl) 
#        s.minimum = str(scl) 
#        s.maximum = str(scl) 
#        s.inclusiveMinimum = True
#        s.inclusiveMaximum = True
#        s.uncertainty = '+{},-{}'.format(errlo,errhi)
#        s.approximate = True
#        return s





"""
Template for developing operations for paws.
"""

import <modulepath.>Operation as op
from <modulepath.>Operation import Operation

# Replace <OperationName> with a SHORT operation title 
class <OperationName>(Operation):
    # Replace <Description of Operation> with a detailed description 
    """<Description of Operation>"""

    def __init__(self):
        """<Description of anything notable performed during __init__()>"""
        # Name the inputs and outputs for your operation.
        input_names = ['<input_name_1>','<input_name_2>',<...>]
        output_names = ['<output_name_1>','<output_name_2>',<...>]
        # Replace <OperationName> with the same name chosen above
        super(<OperationName>,self).__init__(input_names,output_names)
        # Provide a basic description for each of the inputs and outputs 
        self.input_doc['<input_name_1>'] = '<expectations for input 1>'
        self.input_doc['<input_name_2>'] = '<etc>'
        self.output_doc['<output_name_1>'] = '<description of output 1>'
        self.output_doc['<output_name_2>'] = '<etc>'
        # OPTIONAL: set default sources, types, values for the inputs.
        # Valid sources:
        #   op.no_input (default input to None), 
        #   op.wf_input (take input from another operation in the workflow), 
        #   op.text_input (manual text input)
        #   op.plugin_input (take input from a PawsPlugin)
        #   op.batch_input (input provided by a batch/realtime operation)
        self.input_src['<input_name_1>'] = <op.some_source>
        self.input_src['<input_name_2>'] = <etc>
        # Valid types: op.none_type (None), op.str_type (string), 
        #   op.int_type (integer), op.float_type (float), 
        #   op.bool_type (boolean), op.ref_type (direct reference),
        #   op.path_type (a path to something in the filesystem or workflow), 
        #   op.auto_type (default for chosen source, or input set by batch)
        self.input_type['<input_name_1>'] = <op.some_type>
        self.input_type['<input_name_2>'] = <etc>
        
    # Write a run() function for this Operation.
    def run(self):
        """<Description of processing performed by run()>"""
        # Optional- create references in the local namespace for cleaner code.
        <inp1> = self.inputs['<input_name_1>']
        <inp2> = self.inputs['<input_name_2>']
        <etc>
        # Perform the computation
        < ... >
        # Save the output
        self.outputs['<output_name_1>'] = <computed_value_1>
        self.outputs['<output_name_2>'] = <etc>




from functools import partial

from PySide import QtCore, QtGui, QtUiTools

from ..core.operations import Operation as op
from ..core import pawstools
from ..core.models.ListModel import ListModel
from .InputLoader import InputLoader
from . import uitools

class WfUiManager(QtCore.QObject):
    # TODO: remove calls to workflow instance methods.
    # Replace with calls to workflow manager.

    def __init__(self,qwfman,qopman,qplugman):
        ui_file = QtCore.QFile(pawstools.sourcedir+"/ui/qtui/wf_editor.ui")
        ui_file.open(QtCore.QFile.ReadOnly)
        self.ui = QtUiTools.QUiLoader().load(ui_file)
        ui_file.close()
        # set self.ui to be deleted and to signal destroyed()
        # when its window is closed
        self.ui.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.qwfman = qwfman
        self.qopman = qopman 
        self.qplugman = qplugman 
        self.op = None
        self.setup_ui()
        # Dicts to keep track of input widgets, keyed by input variable names
        self.src_widgets = {} 
        self.type_widgets = {} 
        self.val_widgets = {} 
        self.btn_widgets = {} 
        self.input_loaders = {} 
        # Column definitions for the io layout        
        self.name_col = 0
        self.eq_col = 1
        self.src_col = 2
        self.type_col = 3
        self.val_col = 4
        self.btn_col = 5
        self.set_wf()

    def set_wf(self,wf_idx=None):
        if wf_idx is None:
            wf_idx = self.ui.wf_selector.currentIndex()
        wfname = self.ui.wf_selector.model().list_data()[wf_idx]
        self.ui.wf_browser.setModel(self.qwfman.qworkflows[wfname])
        self.ui.wf_browser.setRootIndex(self.qwfman.qworkflows[wfname].root_index())
        self.ui.wf_browser.hideColumn(1)
        self.ui.wf_browser.hideColumn(2)

    def current_wf(self):
        wfname = self.current_wfname()
        if wfname:
            return self.qwfman.qworkflows[wfname]

    def current_wfname(self):
        current_wf_idx = self.ui.wf_selector.currentIndex()
        if current_wf_idx > -1:
            return self.ui.wf_selector.model().list_data()[current_wf_idx]
        else:
            return None

    def get_op(self,src_qtree=None,itm_idx=QtCore.QModelIndex()):
        if src_qtree is None:
            src_qtree = self.current_wf()
        x = src_qtree.get_data_from_index(itm_idx)
        if x is not None:
            try:
                new_op_flag = issubclass(x,op.Operation)
            except:
                new_op_flag = False
            try:
                existing_op_flag = isinstance(x,op.Operation)
            except:
                existing_op_flag = False
            if not new_op_flag and not existing_op_flag:  
                self.op = None
                self.clear_io()
                #self.ui.op_info.setPlainText('Selected item: {}'.format(x))
            elif new_op_flag:
                # Create a new Operation 
                self.create_op(x)
            elif existing_op_flag:
                # Copy the setup of existing Operation
                op_dict = self.qwfman.wfman.op_setup_dict(x)
                # Get a new op with the same setup
                op_copy = self.qwfman.wfman.build_op_from_dict(op_dict,self.qopman.opman)
                op_tag = src_qtree.get_uri_of_index(itm_idx)
                self.set_op(op_copy,op_tag)

    def set_op(self,op,op_tag):
        """Set up ui elements around input op"""
        self.op = op
        #self.ui.op_info.setPlainText(self.op.description())
        self.build_io()
        self.ui.tag_entry.setText(op_tag)

    def create_op(self,op):
        """Instantiate op, call self.set_op()"""
        new_op = op()
        new_op_tag = self.current_wf()._tree.make_unique_uri(type(new_op).__name__)
        new_op.load_defaults()
        self.set_op(new_op,new_op_tag)

    def remove_op(self):
        """
        remove the selected operation from the workflow
        """
        idx = self.ui.wf_browser.currentIndex()
        if idx.isValid():
            itm = self.current_wf().get_from_index(idx) 
            while not itm.parent == self.current_wf()._tree._root_item:
                itm = itm.parent
            self.current_wf().remove_item(itm.tag)
            self.qwfman.wf_updated.emit(self.current_wfname())

    def load_op(self):
        """
        Package the finished Operation, put it in self.current_wf()
        """ 
        # Make sure all inputs are loaded
        for name in self.op.inputs.keys():
            self.set_input(name)
        tag = str(self.ui.tag_entry.text())
        if self.current_wf()._tree.is_tag_valid(tag):
            self.current_wf().set_item(tag,self.op)
            #import pdb; pdb.set_trace()
        else:
            # Request a different tag 
            msg_ui = uitools.message_ui(self.ui)
            msg_ui.setWindowTitle("Tag Error")
            msg_ui.message_box.setPlainText(self.current_wf()._tree.tag_error(tag))
            msg_ui.show()
        self.op = None
        self.clear_io()
        self.qwfman.wf_updated.emit(self.current_wfname())
        #replace_op_flag = self.current_wf().wf.contains_uri(tag)
        #if replace_op_flag: 
        #    self.current_wf().update_op(tag,self.op)
        #    self.op = None
        #else:
        #    self.current_wf().add_op(tag,self.op) 
        #    self.op = None
        #    self.clear_io()

    def set_input(self,name,src_ui=None):
        """
        Call build_input_locator to package an InputLocator for named input. 
        Store it in self.op.input_locator[name].
        This is called on all inputs when an Operation is loaded,
        so it should not alter an identical already-loaded input.
        """
        il = self.build_input_locator(name,src_ui)
        self.op.input_locator[name] = il
        # dereference the old input
        self.op.inputs[name] = None 

    def build_input_locator(self,name,ui=None):
        """
        Create an InputLocator for named input.
        If a ui is provided, it should be an input_loader.ui,
        and this should load data from that ui.
        """
        src = self.src_widgets[name].currentIndex()
        tp = self.type_widgets[name].currentIndex()
        il = None
        if src == op.no_input:
            il = op.InputLocator() 
        elif src == op.batch_input:
            if tp == op.auto_type:
                val = 'batch_input' 
            else:
                val = None
            il = op.InputLocator(src,tp,val) 
        else:
            # all other input sources use a ui for loading
            if ui:
                val = self.fetch_from_input_ui(ui) 
                il = op.InputLocator(src,tp,val)
                self.val_widgets[name].setText(str(il.val))
                ui.close()
                ui.deleteLater()
                # dereference the ui now that it is closed...
                self.input_loaders[name] = None
            else:
                # else, the input has already been loaded.
                # if src, tp, and val match, do nothing.
                val = self.val_widgets[name].text()
                if (self.op.input_locator[name].src == src 
                and self.op.input_locator[name].tp == tp 
                and str(self.op.input_locator[name].val) == val):
                    il = self.op.input_locator[name]
                else:
                    il = op.InputLocator(src,tp,val)
        return il

    def fetch_from_input_ui(self,ui):
        # no matter the input source, 
        # ui.values_list.model().list_data()
        # should be a list of strings.
        val = ui.values_list.model().list_data()
        # if ui.list_toggle is not checked,
        # the value should be unpacked. 
        if not ui.list_toggle.isChecked():
            if len(val) == 0:
                val = None
            else:
                val = val[0]
        return val

    def fetch_data(self,name):
        if name in self.input_loaders.keys():
            if self.input_loaders[name]:
                self.input_loaders[name].ui.close()
                self.input_loaders[name] = None
        src = self.src_widgets[name].currentIndex()
        input_loader_title = self.ui.tag_entry.text()+'.'+op.inputs_tag+'.'+name
        if src == op.wf_input:
            inp_loader = InputLoader(input_loader_title,src,self.qwfman,self.ui)
            inp_loader.ui.wf_selector.setCurrentIndex(self.ui.wf_selector.currentIndex())
            inp_loader.set_wf()
        elif src == op.fs_input:
            trmod = QtGui.QFileSystemModel()
            inp_loader = InputLoader(input_loader_title,src,trmod,self.ui)
        elif src == op.plugin_input:
            inp_loader = InputLoader(input_loader_title,src,self.qplugman,self.ui)
        elif src == op.text_input:
            inp_loader = InputLoader(input_loader_title,src,None,self.ui)
        if self.op.input_locator[name].src == src and self.op.input_locator[name].val is not None:
            if isinstance(self.op.input_locator[name].val,list):
                inp_loader.set_list_toggle()
                #for v in self.op.input_locator[name].val:
                inp_loader.add_values(self.op.input_locator[name].val)
            else:
                inp_loader.add_values([self.op.input_locator[name].val])
        inp_loader.ui.finish_button.clicked.connect( partial(self.set_input,name,inp_loader.ui) )
        self.input_loaders[name] = inp_loader
        #These do not need to be deleted on close. 
        #They will be deleted on clear_io() or when WfUiManager is closed.
        #inp_loader.ui.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        inp_loader.ui.show()

    def clear_io(self):
        self.ui.op_name.setText('')
        self.ui.tag_entry.setText('')
        self.ui.load_button.setEnabled(False)
        # clean up internal objects via deleteLater()
        for nm,widg in (self.src_widgets.items()+self.type_widgets.items()
        +self.val_widgets.items()+self.btn_widgets.items()):
            if widg:
                widg.close()
                widg.deleteLater()
        for nm,il in self.input_loaders.items():
            if il:
                il.ui.close()
                il.ui.deleteLater()
        # dereference them to ensure they are collected
        self.src_widgets = {}
        self.type_widgets = {}
        self.val_widgets = {}
        self.btn_widgets = {}
        self.input_loaders = {}
        n_inp_widgets = self.ui.input_layout.count()
        for i in range(n_inp_widgets-1,-1,-1):
            item = self.ui.input_layout.takeAt(i)
            item.widget().close()
            item.widget().deleteLater()
        n_out_widgets = self.ui.output_layout.count()
        for i in range(n_out_widgets-1,-1,-1):
            item = self.ui.output_layout.takeAt(i)
            item.widget().close()
            item.widget().deleteLater()
    
    def reset_input_headers(self):
        if len(self.op.inputs) > 0:
            self.ui.input_layout.addWidget(uitools.r_hdr_widget('name'),0,self.name_col,1,1)
            self.ui.input_layout.addWidget(uitools.hdr_widget('source'),0,self.src_col,1,1)
            self.ui.input_layout.addWidget(uitools.hdr_widget('type'),0,self.type_col,1,1)
            self.ui.input_layout.addWidget(uitools.hdr_widget('value'),0,self.val_col,1,1)

    def reset_output_headers(self):
        if len(self.op.outputs) > 0:
            self.ui.output_layout.addWidget(uitools.r_hdr_widget('name'),0,self.name_col,1,1)
            self.ui.output_layout.addWidget(uitools.hdr_widget('description'),0,self.src_col,1,self.btn_col-self.src_col)

    def build_io(self):
        self.clear_io()
        self.reset_input_headers()
        self.reset_output_headers()
        self.ui.op_name.setText(type(self.op).__name__)
        self.ui.load_button.setEnabled(True)
        for i,name in zip(range(1,len(self.op.inputs)+1),self.op.inputs.keys()):
            # name 
            name_widget = uitools.name_widget(name)
            self.ui.input_layout.addWidget( name_widget,i,self.name_col,1,1 )
            # '=' 
            eq_widget = uitools.smalltext_widget('=')
            self.ui.input_layout.addWidget(eq_widget,i,self.eq_col,1,1)
            # source 
            src_widget = uitools.src_selection_widget()
            if self.op.input_locator[name] is not None:
                src = self.op.input_locator[name].src
            else:
                src = op.no_input 
            src_widget.setCurrentIndex(src)
            self.ui.input_layout.addWidget( src_widget,i,self.src_col,1,1 )
            self.src_widgets[name] = src_widget 
            self.src_widgets[name].currentIndexChanged.connect( partial(self.reset_type_widget,name,i) )
            self.reset_type_widget(name,i,src)
        # Now handle outputs, much easier
        for i,name in zip(range(1,len(self.op.outputs)+1),self.op.outputs.keys()):
            # TODO: allow multiple lines for the desc_widgets.
            name_widget = uitools.name_widget(name)
            self.ui.output_layout.addWidget(name_widget,i,self.name_col)
            eq_widget = uitools.smalltext_widget('=')
            self.ui.output_layout.addWidget(eq_widget,i,self.eq_col)
            if self.op.output_doc[name]:
                desc_widget = uitools.bigtext_widget(self.op.output_doc[name])
            else:
                desc_widget = uitools.bigtext_widget('No description found.')
            self.ui.output_layout.addWidget(desc_widget,i,self.src_col,1,self.btn_col-self.src_col)

    def reset_type_widget(self,name,row,src=None):
        if name in self.type_widgets.keys():
            if self.type_widgets[name]:
                self.type_widgets[name].close()
        if not src:
            src = self.src_widgets[name].currentIndex()
        new_type_widget = uitools.type_selection_widget(src)
        if not self.op.input_locator[name].src == op.no_input:
            if (self.op.input_locator[name].tp not in op.invalid_types[src]
            and self.op.input_locator[name].src == src):
                # load operation defaults
                new_type_widget.setCurrentIndex(self.op.input_locator[name].tp)
            else:
                # set sensible defaults
                if src == op.fs_input:
                    new_type_widget.setCurrentIndex(op.path_type)
                elif src in [op.wf_input,op.plugin_input]:
                    new_type_widget.setCurrentIndex(op.ref_type)
                elif src == op.batch_input:
                    new_type_widget.setCurrentIndex(op.auto_type)
        new_type_widget.currentIndexChanged.connect( partial(self.reset_val_widget,name,row,src) )            
        self.type_widgets[name] = new_type_widget  
        self.ui.input_layout.addWidget(new_type_widget,row,self.type_col,1,1)
        tp = self.type_widgets[name].currentIndex()
        self.reset_val_widget(name,row,src,tp)

    def reset_val_widget(self,name,row,src=None,tp=None):
        # TODO: Make val widgets expand height according to their content 
        if name in self.val_widgets.keys():
            if self.val_widgets[name]:
                self.val_widgets[name].close()
        if name in self.btn_widgets.keys():
            if self.btn_widgets[name]:
                self.btn_widgets[name].close()
        if name in self.input_loaders.keys():
            if self.input_loaders[name]:
                self.input_loaders[name].ui.close()
                self.input_loaders[name] = None
        if not src:
            src = self.src_widgets[name].currentIndex()
        if not tp:
            tp = self.type_widgets[name].currentIndex()
        btn_widget = QtGui.QPushButton()
        val_widget = QtGui.QLineEdit()
        #val_widget = QtGui.QTextEdit()
        if src == op.no_input or tp == op.none_type: 
            btn_widget.setText('no input')
            btn_widget.setEnabled(False)
            val_widget.setText('None')
        elif src == op.batch_input:
            btn_widget.setText('auto')
            btn_widget.setEnabled(False)
            val_widget.setText('auto (batch)')
        elif src in [op.wf_input,op.fs_input,op.plugin_input,op.text_input]:
            btn_widget.setText('edit...')
            btn_widget.clicked.connect( partial(self.fetch_data,name) )
            if self.op.input_locator[name]:
                if self.op.input_locator[name].src == src:# and self.op.input_locator[name].tp == tp:
                    val_widget.setText(str(self.op.input_locator[name].val))
        val_widget.setReadOnly(True)
        self.ui.input_layout.addWidget(val_widget,row,self.val_col,1,1)
        self.ui.input_layout.addWidget(btn_widget,row,self.btn_col,1,1)
        self.val_widgets[name] = val_widget
        self.btn_widgets[name] = btn_widget

    def setup_ui(self):
        # LABELS and STYLES
        self.ui.setStyleSheet( "QLineEdit { border: none }" + self.ui.styleSheet() )
        self.ui.setWindowTitle("workflow setup")
        self.ui.input_box.setTitle("INPUTS")
        self.ui.output_box.setTitle("OUTPUTS")
        self.ui.finish_box.setTitle("FINISH / LOAD")
        self.ui.wf_box.setTitle("WORKFLOWS")
        self.ui.op_box.setTitle("OPERATIONS")
        self.ui.tag_prompt.setText('operation tag:')
        self.ui.tag_prompt.setReadOnly(True)
        self.ui.op_name.setText('-select an operation to begin setup-')
        self.ui.op_name.setAlignment(QtCore.Qt.AlignCenter)
        self.ui.tag_prompt.setAlignment(QtCore.Qt.AlignRight)
        self.ui.tag_prompt.setStyleSheet( "QLineEdit { background-color: transparent }" 
        + self.ui.tag_prompt.styleSheet() )
        self.ui.load_button.setText("&Finish")
        self.ui.rm_op_button.setText("&Remove selected operation")
        # SIGNALS, SLOTS, MODELS, VIEWS 
        lm = ListModel(self.qwfman.qworkflows.keys())
        self.ui.wf_selector.setModel(lm)
        self.ui.wf_selector.currentIndexChanged.connect(self.set_wf)
        #self.ui.wf_selector.currentIndexChanged.connect( partial(self.set_wf) )
        self.ui.wf_selector.activated.connect(self.set_wf)
        self.ui.wf_browser.clicked.connect( partial(self.get_op,None) )
        self.ui.rm_op_button.clicked.connect(self.remove_op)
        self.ui.op_selector.setModel(self.qopman)
        self.ui.op_selector.hideColumn(1)
        self.ui.op_selector.hideColumn(2)
        self.ui.op_selector.clicked.connect( partial(self.get_op,self.qopman) )
        self.ui.op_selector.clicked.connect( partial(uitools.toggle_expand,self.ui.op_selector) ) 
        self.ui.op_selector.setRootIndex(self.qopman.root_index())
        self.ui.load_button.clicked.connect(self.load_op)
        self.ui.load_button.setDefault(True)
        self.ui.load_button.setEnabled(False)
        # LAYOUT AND SIZING 
        szp = QtGui.QSizePolicy()
        szp.setHorizontalPolicy(QtGui.QSizePolicy.Preferred)
        szp.setVerticalPolicy(QtGui.QSizePolicy.Preferred)
        szp.setHorizontalStretch(3)
        szp.setVerticalStretch(1)
        self.ui.input_frame.setSizePolicy(szp)
        finish_box_szp = QtGui.QSizePolicy()
        finish_box_szp.setHorizontalPolicy(QtGui.QSizePolicy.Preferred)
        finish_box_szp.setVerticalPolicy(QtGui.QSizePolicy.Maximum)
        self.ui.finish_box.setSizePolicy(finish_box_szp)
        io_box_szp = QtGui.QSizePolicy()
        io_box_szp.setHorizontalPolicy(QtGui.QSizePolicy.Preferred)
        io_box_szp.setVerticalPolicy(QtGui.QSizePolicy.Maximum)
        self.ui.input_box.setSizePolicy(io_box_szp)
        self.ui.output_box.setSizePolicy(io_box_szp)
        self.ui.load_button.setMinimumWidth(100)
        self.ui.tag_prompt.setMaximumWidth(150)




"""
Operations config and processing routines
"""

import glob
import copy
from collections import Iterator
from collections import OrderedDict

from . import Operation as op 

class FileSystemIterator(Iterator):

    def __init__(self,dirpath,regex,include_existing_files=True):
        self.dirpath = dirpath
        self.rx = regex
        self.paths_done = []
        if not include_existing_files:
            self.paths_done = glob.glob(self.dirpath+'/'+self.rx)
        super(FileSystemIterator,self).__init__()

    def next(self):
        batch_list = glob.glob(self.dirpath+'/'+self.rx)
        for path in batch_list:
            if not path in self.paths_done:
                self.paths_done.append(path)
                return [path]
        return [None]

class ExecutionError(Exception):
    def __init__(self,msg):
        super(ExecutionError,self).__init__(self,msg)

def get_uri_from_dict(uri,d):
    keys = uri.split('.')
    itm = d
    for k in keys:
        if not isinstance(itm,dict):
            msg = 'something in {} is not a dict'.format(uri)
            raise KeyError(msg)
        if not k in itm.keys():
            msg = 'did not find uri {} in dict'.format(uri)
            raise KeyError(msg)
        else:
            itm = itm[k]
    return itm

def dict_contains_uri(uri,d):
    keys = uri.split('.')
    itm = d
    for k in keys:
        if not k in itm.keys():
            return False
        else:
            itm = itm[k]
    return True


####### functions for fetching inputs and loading them into operations #######

def cast_type_val(tp,val):
    """
    Perform type casting for operation inputs.
    This should be called only for source = text_input.
    """
    if tp == op.none_type:
        val = None 
    elif tp == op.int_type:
        val = int(val)
    elif tp == op.float_type:
        val = float(val)
    elif tp == op.str_type:
        val = str(val)
    elif tp == op.bool_type:
        val = bool(eval(str(val)))
    else:
        msg = 'type selection {}, should be one of {}'.format(tp,op.valid_types)
        raise ValueError(msg)
    return val

def locate_input(il,wf=None,plugin_manager=None):
    """
    Return the data pointed to by a given InputLocator object.
    A Workflow and a PluginManager can be provided 
    as optional arguments,
    in which case they are used to fetch data.
    """
    if il.src == op.no_input or il.tp == op.none_type:
        return None
    elif il.src == op.batch_input:
        # Expect this input to have been set by Workflow Manager.
        return il.data 
    elif il.src == op.text_input: 
        if isinstance(il.val,list):
            return [cast_type_val(il.tp,v) for v in il.val]
        else:
            return cast_type_val(il.tp,il.val)
    elif il.src == op.wf_input:
        if il.tp == op.ref_type:

            # Note, this will return whatever data is stored in the TreeItem at uri.
            # If il.val is the uri of an input that has not yet been loaded,
            # this means it will get the InputLocator that currently inhabits that uri.

            # Note, this problem has now been fixed by changing Workflow.build_dict()
            # to not substitute InputLocators for inputs that had not been loaded.

            if isinstance(il.val,list):
                return [wf.get_data_from_uri(v) for v in il.val]
            else:
                return wf.get_data_from_uri(il.val)
        elif il.tp == op.path_type: 
            if isinstance(il.val,list):
                return [str(v) for v in il.val]
            else:
                return str(il.val)
    elif il.src == op.plugin_input:
        if il.tp == op.ref_type:
            if isinstance(il.val,list):
                return [plugin_manager.get_data_from_uri(v) for v in il.val]
            elif il.val is not None:
                return plugin_manager.get_data_from_uri(il.val)
            else:
                return None
        elif il.tp == op.path_type:
            if isinstance(il.val,list):
                return [str(v) for v in il.val]
            else:
                return str(il.val)
    elif il.src == op.fs_input:
        if isinstance(il.val,list):
            return [str(v) for v in il.val]
        else:
            return str(il.val)
    else: 
        msg = 'found input source {}, should be one of {}'.format(
        il.src, op.valid_sources)
        raise ValueError(msg)

def load_inputs(operation,wf=None,plugin_manager=None):
    """
    Loads input data for an Operation from its input_locators.
    A Workflow and a PluginManager can be provided 
    as optional arguments,
    in which case they are used to fetch data.
    """
    for name,il in operation.input_locator.items():
        if isinstance(il,op.InputLocator):
            il.data = locate_input(il,wf,plugin_manager)
            operation.inputs[name] = il.data
        else:
            msg = '[{}] Found broken Operation.input_locator for {}: {}'.format(
            __name__, name, il)
            raise ValueError(msg)

####### functions having to do with workflow execution #######
# TODO: consider creating a wftools module instead

def get_valid_wf_inputs(op_tag,operation):
    """
    Return the TreeModel uris of the op and its inputs/outputs 
    that are eligible as downstream inputs in the workflow.
    """
    # valid_wf_inputs should be the operation, its input and output dicts, and their respective entries
    valid_wf_inputs = [op_tag,op_tag+'.'+op.inputs_tag,op_tag+'.'+op.outputs_tag]
    valid_wf_inputs += [op_tag+'.'+op.outputs_tag+'.'+k for k in operation.outputs.keys()]
    valid_wf_inputs += [op_tag+'.'+op.inputs_tag+'.'+k for k in operation.inputs.keys()]
    return valid_wf_inputs
    
def stack_size(stk):
    sz = 0
    for lst in stk:
        for lst_itm in lst:
            if isinstance(lst_itm,list):
                sz += stack_size(lst_itm)
            else:
                sz += 1
    return sz

def stack_contains(itm,stk):
    for lst in stk:
        if itm in lst:
            return True
        for lst_itm in lst:
            if isinstance(lst_itm,list):
                if stack_contains(itm,lst_itm):
                    return True
    return False

def print_stack(stk):
    stktxt = ''
    opt_newline = '\n'
    for i,lst in zip(range(len(stk)),stk):
        if i == len(stk)-1:
            opt_newline = ''
        if len(lst) > 1:
            if isinstance(lst[1],list):
                substk = lst[1]
                stktxt += ('[\'{}\':\n{}\n]'+opt_newline).format(lst[0],print_stack(lst[1]))
            else:
                stktxt += ('{}'+opt_newline).format(lst)
        else:
            stktxt += ('{}'+opt_newline).format(lst)
    return stktxt

def is_op_ready(wf,op_tag,valid_wf_inputs,batch_routes=[]):
    operation = wf.get_data_from_uri(op_tag)
    if operation._batch_flag: 
        b_stk,op_rdy,diagnostics = batch_op_stack(wf,op_tag,valid_wf_inputs)
    elif operation._realtime_flag: 
        rt_stk,op_rdy,diagnostics = batch_op_stack(wf,op_tag,valid_wf_inputs)
    else:
        inputs_rdy = []
        diagnostics = {} 
        for name,il in operation.input_locator.items():
            msg = ''
            if (il.src == op.wf_input and il.tp == op.ref_type 
            and not il.val in valid_wf_inputs):
                inp_rdy = False
                msg = str('Operation input {}.inputs.{} (={}) '.format(op_tag,name,il.val)
                + 'not found in valid Workflow input list: {}'.format(valid_wf_inputs))
            elif (il.src == op.batch_input 
            and not op_tag+'.'+op.inputs_tag+'.'+name in batch_routes):
                inp_rdy = False
                msg = str('Operation input {}.inputs.{} (={}) '.format(op_tag,name,il.val)
                + 'expects batch input but is not listed in batch routes: {}'.format(batch_routes))
            else:
                inp_rdy = True
            inputs_rdy.append(inp_rdy)
            diagnostics[op_tag+'.'+op.inputs_tag+'.'+name] = msg
        if all(inputs_rdy):
            op_rdy = True
        else:
            op_rdy = False
    return op_rdy,diagnostics 

def batch_op_stack(wf,batch_op_tag,valid_wf_inputs):
    """
    Use batch_op.batch_ops() and a list of valid_wf_inputs 
    to build a stack (list) of lists of operations suitable for serial execution.
    """
    batch_op = wf.get_data_from_uri(batch_op_tag)
    # Batch and Realtime execution operations 
    # expect to have two inputs loaded by internal methods 
    # before calling batch_op.batch_ops() and batch_op.input_routes()
    batch_op.set_batch_ops(wf)
    batch_op.set_input_routes(wf)
    op_tags = batch_op.batch_ops()
    #load_inputs(batch_op,wf,plugin_manager)
    #op_tags = []
    #if batch_op._realtime_flag: 
    #    op_tags = batch_op.realtime_ops()
    #elif batch_op._batch_flag: 
    #    op_tags = batch_op.batch_ops()
    # make a copy of valid_wf_inputs
    # so that the existing valid_wf_inputs list is not mutated 
    valid_batch_inputs = copy.copy(valid_wf_inputs)
    # add the batch's own valid inputs to the list
    valid_batch_inputs += get_valid_wf_inputs(batch_op_tag,batch_op)
    # build the batch substack
    b_stk = []
    layer = []
    diagnostics = {}
    for op_tag in op_tags:
        op_rdy,op_diag = is_op_ready(wf,op_tag,valid_batch_inputs,batch_op.input_routes())
        diagnostics.update(op_diag)
        if op_rdy:
            layer.append(op_tag)
    while any(layer):
        b_stk.append(layer)
        for op_tag in layer:
            operation = wf.get_data_from_uri(op_tag)
            valid_batch_inputs += get_valid_wf_inputs(op_tag,operation)
        layer = []
        for op_tag in op_tags:
            op_rdy,op_diag = is_op_ready(wf,op_tag,valid_batch_inputs,batch_op.input_routes())
            diagnostics.update(op_diag)
            if op_rdy and not stack_contains(op_tag,b_stk):
                layer.append(op_tag)
    b_rdy = len(op_tags) == stack_size(b_stk) 
    return b_stk,b_rdy,diagnostics 

# TODO: the following
def check_wf(wf):
    """
    Check the dependencies of the workflow.
    Ensure that all loaded operations have inputs that make sense.
    Return a status code and message for each of the Operations.
    """
    pass

def execution_stack(wf):
    """
    Build a stack (list) of lists of Operation uris,
    such that each list indicates a set of Operations
    whose dependencies are satisfied by the Operations above them.
    For Batch or Realtime operations, 
    the layer should be of the form[batch_name,[batch_stack]],
    where batch_name indicates the batch controller Operation,
    and batch_stack is built from batch_op_stack().
    """
    stk = []
    valid_wf_inputs = []
    diagnostics = {}
    continue_flag = True
    while not stack_size(stk) == wf.n_ops() and continue_flag:
        ops_rdy = []
        ops_not_rdy = []
        for itm in wf._root_item.children:
            if not stack_contains(itm.tag,stk):
                op_rdy,op_diag = is_op_ready(wf,itm.tag,valid_wf_inputs)
                diagnostics.update(op_diag)
                if op_rdy:
                    ops_rdy.append(itm.tag)
                else:
                    ops_not_rdy.append(itm.tag)
        # Finished building list of ops currently ready. Now filter these into stack.
        if any(ops_rdy):
            # Which of these are not Batch/Realtime ops?
            non_batch_rdy = []
            for op_tag in ops_rdy:
                operation = wf.get_data_from_uri(op_tag)
                if not any([operation._batch_flag,operation._realtime_flag]):
                    non_batch_rdy.append(op_tag)
            if any(non_batch_rdy):
                ops_rdy = non_batch_rdy
                stk.append(ops_rdy)
                for op_tag in ops_rdy:
                    operation = wf.get_data_from_uri(op_tag)
                    valid_wf_inputs += get_valid_wf_inputs(op_tag,operation)
            else:
                batch_tag = ops_rdy[0]
                ops_rdy = [batch_tag]
                batch_op = wf.get_data_from_uri(batch_tag)
                batch_stk,batch_rdy,batch_diag = batch_op_stack(
                wf,batch_tag,valid_wf_inputs)
                diagnostics.update(batch_diag)
                stk.append([batch_tag,batch_stk])
                valid_wf_inputs += get_valid_wf_inputs(batch_tag,batch_op)
        else:
            continue_flag = False
    return stk,diagnostics





from PySide import QtCore

from ..core.models.TreeModel import TreeModel

class QTreeModel(QtCore.QAbstractItemModel):
    """
    A Qt Model-View interface for a TreeModel.
    Required virtual methods: index(), parent(), rowCount(), columnCount(), and data().
    Resizeable TreeModels should implement
    insertRows(), removeRows(), insertColumns(), and removeColumns().
    To customize the header in QAbstractItemViews, implement headerData().
    """

    def __init__(self,trmod=TreeModel()):
        super(QTreeModel,self).__init__()
        self._tree=TreeModel()
        if isinstance(trmod,TreeModel):
            for c_tag in trmod.root_tags():
                self.set_item(c_tag,trmod[c_tag])
            self._tree=trmod

    # Subclass of QAbstractItemModel must implement index()
    def index(self,row,col,p_idx):
        """
        Returns QModelIndex address of int row, int col, under QModelIndex p_idx.
        If a row, column, p_idx combination is invalid, return QModelIndex().
        """
        if not p_idx.isValid():
            if row == 0:
                return self.root_index()
            else:
                return QtCore.QModelIndex()
        else:
            p_itm = self.get_from_index(p_idx)
            if row < p_itm.n_children() and row >= 0:
                return self.createIndex(row,col,p_itm.children[row])
            else:
                return QtCore.QModelIndex()

    def root_index(self):
        return self.createIndex(0,0,self._tree._root_item)

    def root_item(self):
        return self._tree._root_item

    def get_from_index(self,idx=QtCore.QModelIndex()):
        """
        For a valid QModelIndex, return idx.internalPointer().
        For an invalid index, return None.
        """
        if idx.isValid():
            return idx.internalPointer() 
        else:
            return None 

    def get_data_from_index(self,idx=QtCore.QModelIndex()):
        uri = self.get_uri_of_index(idx)
        return self._tree.get_data_from_uri(uri)

    def get_uri_of_index(self,idx):
        itm = self.get_from_index(idx)
        return self._tree.build_uri(itm)

    def get_index_of_uri(self,uri):
        path = uri.split('.')
        idx = self.root_index()
        for k in path[:-1]:
            itm = self.get_from_index(idx)
            child_keys = [c.tag for c in itm.children]
            idx = self.index(child_keys.index(k),0,idx)
        k = path[-1]
        if k:
            itm = self.get_from_index(idx)
            child_keys = [c.tag for c in itm.children]
            return self.index(child_keys.index(k),0,idx)
        elif k == '':
            return idx 

    def n_children(self,parent_idx=None):
        if parent_idx is None:
            parent_idx = self.root_index()
        return self.rowCount(parent_idx)

    def set_item_at_uri(self,itm_uri,itm_data):
        if '.' in itm_uri:
            parent_uri = itm_uri[:itm_uri.rfind('.')]
            #if self._tree.contains_uri(parent_uri):
            parent_idx = self.get_index_of_uri(parent_uri)
            #else:
            #    parent_idx = self.build_to_uri(parent_uri)
            #parent_idx = self.get_index_of_uri(parent_uri)
            itm_tag = itm_uri[itm_uri.rfind('.')+1:]
        else:
            parent_idx = self.root_index()
            itm_tag = itm_uri
        self.set_item(itm_tag,itm_data,parent_idx)        

    #def build_to_uri(self,uri=''):
    #    idx = self.root_index()
    #    if '.' in uri:
    #        parent_uri = uri[:uri.rfind('.')]
    #        if self._tree.contains_uri(parent_uri):
    #            idx = self.get_index_of_uri(parent_uri)
    #        else:
    #            idx = self.build_to_uri(parent_uri)
    #        k = uri.split('.')[-1]
    #        if k == '':
    #            return idx 
    #        elif k is not None:
    #            self.set_item(k,None,idx)
    #            return self.get_index_of_uri(uri) 

    def set_item(self,itm_tag,itm_data=None,parent_idx=None):
        if parent_idx is None:
            parent_idx = self.root_index()
        parent_itm = self.get_from_index(parent_idx)
        parent_uri = self.get_uri_of_index(parent_idx)
        if parent_uri:
            itm_uri = parent_uri+'.'+itm_tag
        else:
            itm_uri = itm_tag
        # add TreeItems to index the new model content:
        # have to implement this differently than TreeModel
        # to inject calls to beginInsertRows and endInsertRows.        
        treedata = self._tree.build_tree(itm_data)
        self.tree_update(parent_idx,itm_tag,treedata)
        # store the data:
        # self._tree.tree_update() will be called with no effect. 
        self._tree.set_item(itm_uri,itm_data)
        #if isinstance(treedata,dict):
        #    itm_idx = self.get_index_of_uri(itm_uri)
        #    # also store any children
        #    for k,val in treedata.items():
        #        child_uri = itm_uri+'.'+k
        #        if not self._tree.contains_uri(child_uri):
        #            if isinstance(itm_data,list):
        #                self.set_item(k,itm_data[int(k)],itm_idx)
        #            else:
        #                # Note- parent item data needs to implement __getitem__
        #                self.set_item(k,itm_data[k],itm_idx) 
        # signal dataChanged down the tree
        child_keys = [c.tag for c in parent_itm.children]
        itm_idx = self.index(child_keys.index(itm_tag),0,parent_idx)
        self.tree_dataChanged(itm_idx) 

    def tree_update_at_uri(self,itm_uri,itm_data):
        if '.' in itm_uri:
            parent_uri = itm_uri[:itm_uri.rfind('.')]
            parent_idx = self.get_index_of_uri(parent_uri)
            itm_tag = itm_uri[itm_uri.rfind('.')+1:]
        else:
            parent_idx = self.root_index()
            itm_tag = itm_uri
        self.tree_update(parent_idx,itm_tag,self._tree.build_tree(itm_data))        

    def tree_update(self,parent_idx,itm_tag,itm_tree):
        parent_itm = self.get_from_index(parent_idx)
        child_keys = [c.tag for c in parent_itm.children]
        if itm_tag in child_keys:
            # Find existing itm under parent.
            itm_row = child_keys.index(itm_tag)
            itm = parent_itm.children[itm_row]
            idx = self.index(itm_row,0,parent_idx)
            # Remove any children that do not represent the new itm_tree data
            new_keys = []
            if isinstance(itm_tree,dict):
                new_keys = itm_tree.keys()
            for gc_row in range(itm.n_children())[::-1]:
                gc_itm = itm.children[gc_row]
                if not gc_itm.tag in new_keys:
                    #rm_row = [gc.tag for gc in itm.children].index(grandchild_itm.tag)
                    self.beginRemoveRows(idx,gc_row,gc_row)
                    itm.children.pop(gc_row) 
                    self.endRemoveRows()
        else:
            # Else, put a new TreeItem at the end row
            itm_row = self.n_children(parent_idx)
            itm = self._tree.create_tree_item(parent_itm,itm_tag)
            self.beginInsertRows(parent_idx,itm_row,itm_row)
            parent_itm.children.insert(itm_row,itm)
            self.endInsertRows()
        # If needed, get the index of the new item and recurse 
        if isinstance(itm_tree,dict):
            itm_idx = self.index(itm_row,0,parent_idx)
            for tag,val in itm_tree.items():
                #self.tree_update(itm_idx,tag,self._tree.build_tree(val))
                self.tree_update(itm_idx,tag,val)

    def remove_item(self,itm_tag,parent_idx=None):
        if parent_idx is None:
            parent_idx = self.root_index()
        parent_itm = self.get_from_index(parent_idx)
        child_keys = [c.tag for c in parent_itm.children]
        rm_row = child_keys.index(itm_tag)
        rm_idx = self.index(rm_row,0,parent_idx)
        rm_itm = self.get_from_index(rm_idx) 
        rm_uri = self._tree.build_uri(rm_itm)
        self.beginRemoveRows(parent_idx,rm_row,rm_row)
        self._tree.remove_item(rm_uri)
        #parent_itm.children.pop(rm_row)
        self.endRemoveRows()
        #self.dataChanged.emit(rm_idx,rm_idx)
        self.tree_dataChanged(parent_idx)

    def tree_dataChanged(self,idx):
        itm = idx.internalPointer()
        self.dataChanged.emit(idx,idx)
        for c_row in range(itm.n_children())[::-1]:
            c_idx = self.index(c_row,0,idx)
            self.tree_dataChanged(c_idx)

    # Subclass of QAbstractItemModel must implement parent()
    def parent(self,idx):
        """
        Returns QModelIndex of parent of item at QModelIndex index
        """
        itm = self.get_from_index(idx)
        if itm == self._tree._root_item:
            return QtCore.QModelIndex()
        parent_itm = itm.parent
        if parent_itm == self._tree._root_item:
            return self.root_index()
        parent_tag = parent_itm.tag
        grandparent_itm = parent_itm.parent
        parent_keys = [p.tag for p in grandparent_itm.children]
        parent_row = parent_keys.index(parent_tag)
        return self.createIndex(parent_row,0,grandparent_itm.children[parent_row])
        
    # Subclass of QAbstractItemModel must implement rowCount()
    def rowCount(self,parent_idx=QtCore.QModelIndex()):
        """
        Either give the number of top-level items,
        or count the children of parent
        """
        if not parent_idx.isValid():
            # there is only 1 root item 
            return 1
        else:
            # count children of parent item
            parent_itm = self.get_from_index(parent_idx)
            return parent_itm.n_children()
    
    # Subclass of QAbstractItemModel must implement columnCount()
    def columnCount(self,parent=QtCore.QModelIndex()):
        """
        TreeModels by default have one column.
        More columns can be added by reimplementing columnCount()
        and then providing for them in TreeModel.data().
        """
        return 1

    # QAbstractItemModel subclass must implement 
    # data(QModelIndex[,role=Qt.DisplayRole])
    def data(self,itm_idx,data_role):
        """
        TreeModel's implementation of data() 
        returns the tag of the TreeItem at itm_idx.
        This is only if itm_idx.column() == 0.
        Subclasses can reimplement data() 
        to provide meaningful output for other columns,
        and may consider falling back on super().data() 
        if itm_idx.column() == 0.
        """
        if (not itm_idx.isValid()):
            return None
        #if itm_idx == self.root_index():
        #    return None
        #itm = itm_idx.internalPointer()
        if ((data_role == QtCore.Qt.DisplayRole
        or data_role == QtCore.Qt.ToolTipRole 
        or data_role == QtCore.Qt.StatusTipRole
        or data_role == QtCore.Qt.WhatsThisRole)
        and itm_idx.column() == 0):
            itm = self.get_from_index(itm_idx)
            return itm.tag
        else:
            return None

    def headerData(self,section,orientation,data_role):
        # note: section indicates row or column number, depending on orientation
        if (data_role == QtCore.Qt.DisplayRole and section == 0):
            return "{} item(s)".format(self.n_children())
        else:
            return None




import unittest

class TestAPI(unittest.TestCase):

    def test_start_stop(self):
        # smoke testing paws api
        import paws.api
        paw = paws.api.start()
        self.assertIsInstance(paw,paws.api.PawsAPI)

if __name__ == '__main__':
    unittest.main()




from xicam.plugins import base
import paws.api 
import paws.ui
import paws.ui.ui_manager

class PawsXicamPlugin(base.plugin):
    # The display name in the xi-cam plugin bar
    name = 'Workflow'

    def __init__(self, *args, **kwargs):
        app = paws.ui.ui_app()
        # start core objects
        xipaw = paws.api.start()
        # start ui objects
        ui_manager = paws.ui.ui_manager.UiManager(
        xipaw.op_manager(),xipaw.wf_manager(),xipaw.plugin_manager())

        # Set the widgets in base.plugin containers
        self.centerwidget = ui_manager.ui.center_frame
        self.leftwidget = ui_manager.ui.left_frame
        self.rightwidget = ui_manager.ui.right_frame
        self.bottomwidget = ui_manager.ui.message_board

        # There seems to be a problem with this plugin loading approach,
        # where the frames sometimes fail 
        # to bring their children with them.
        # Adding calls to findChildren() 
        # seems to force the frames
        # to include their children.
        # -LAP
        ui_manager.ui.left_frame.findChildren(object)
        ui_manager.ui.right_frame.findChildren(object)
        ui_manager.ui.center_frame.findChildren(object)

        super(PawsXicamPlugin, self).__init__(*args, **kwargs)




from __future__ import print_function
import inspect
from functools import partial

from PySide import QtCore, QtGui, QtUiTools

from . import uitools
from ..core import pawstools
from ..core.operations.Operation import Operation

class OpUiManager(QtCore.QObject):

    def __init__(self,qopman):
        ui_file = QtCore.QFile(pawstools.sourcedir+"/ui/qtui/op_editor.ui")
        ui_file.open(QtCore.QFile.ReadOnly)
        self.ui = QtUiTools.QUiLoader().load(ui_file)
        ui_file.close()
        # set self.ui to be deleted and to emit destroyed() signal when its window is closed
        self.ui.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.qopman = qopman 
        self.setup_ui()
        self.op = None
        self.filepath = ''
        self.opname_widget = None
        self.path_widget = None
        self.run_widget = None
        self.import_widgets = []
        self.name_widgets = {}
        self.src_widgets = {}
        self.type_widgets = {}
        self.val_widgets = {}
        self.outname_widgets = {}
        self.outdesc_widgets = {}
        self._edit_mode = False

    def get_op(self,idx):
        x = self.qopman.get_data_from_index(idx)
        try:
            op_flag = issubclass(x,Operation)
        except:
            op_flag = False
        if op_flag:
            op_uri = self.qopman.opman.get_uri_of_index(idx)
            self.filepath = pawstools.sourcedir + '/core/operations'
            for uri_piece in op_uri.split('.'):
                self.filepath = self.filepath + '/' + uri_piece
            self.filepath = self.filepath + '.py'
            self.clear_op()
            self.set_op(x())
        else:
            self.clear_op()
            self.ui.op_info.setPlainText('Selected item: {}'.format(x))
    
    def clear_op(self):
        #self.ui.op_name.setText('')
        n_widg = self.ui.op_source_layout.count()
        for i in range(n_widg-1,-1,-1):
            itm = self.ui.op_source_layout.takeAt(i)
            itm.widget().close()
            itm.widget().deleteLater()
        self.name_widgets = {}
        self.src_widgets = {}
        self.type_widgets = {}
        self.val_widgets = {}
        self.outname_widgets = {}
        self.outdesc_widgets = {}

    def set_op(self,op):
        """Set up ui elements around existing input op"""
        self.op = op
        self.op.load_defaults()
        self.ui.op_info.setPlainText(self.op.description())
        self.build_source()

    def build_source(self):
        src_font = QtGui.QFont("Courier New", 10)
        self.opname_widget = QtGui.QLineEdit()
        self.opname_widget.setAlignment(QtCore.Qt.AlignCenter)
        self.opname_widget.setText(type(self.op).__name__)
        # if op is not enabled, display a message
        # else, inspect the op
        path_box = QtGui.QGroupBox('PATH')
        path_lo = QtGui.QGridLayout(path_box)
        self.path_widget = QtGui.QTextEdit()
        self.path_widget.setText(self.filepath)
        self.path_widget.setWordWrapMode(QtGui.QTextOption.NoWrap)
        self.path_widget.setMaximumHeight(50)
        # TODO: Make so that name widget is updated when path widget is edited
        path_lo.addWidget(self.path_widget,0,0,1,1)
        #self.path_widget.setSizePolicy(QtGui.QSizePolicy.Fixed,QtGui.QSizePolicy.MinimumExpanding)
        imports_box = QtGui.QGroupBox('IMPORTS',self.ui)
        imports_lo = QtGui.QGridLayout(imports_box)
        imp_lines = []
        for line in open(self.filepath,'r').readlines():
            if line.find('import') >= 0:
                imp_lines.append(line)
        for ii in range(len(imp_lines)):
            # TODO: try to separate imports: standard, 3rd-party, local
            imp_widget = QtGui.QTextEdit()
            imp_widget.setCurrentFont(src_font)
            imp_widget.setMaximumHeight(30)
            #imp_widget.setSizePolicy(QtGui.QSizePolicy.Preferred,QtGui.QSizePolicy.Fixed)
            imp_widget.setText(imp_lines[ii].strip())
            imports_lo.addWidget(imp_widget,ii,0,1,1)
            self.import_widgets.append(imp_widget)
        #mod_src = inspect.getsource(self.op)
        # set widgets for operation inputs
        inputs_box = QtGui.QGroupBox('INPUTS',self.ui)
        inputs_lo = QtGui.QGridLayout(inputs_box)
        if len(self.op.inputs) > 0:
            inputs_lo.addWidget(uitools.r_hdr_widget('name'),0,0,1,1)
            inputs_lo.addWidget(uitools.hdr_widget('default source'),0,2,1,1)
            inputs_lo.addWidget(uitools.hdr_widget('default type'),0,3,1,1)
            inputs_lo.addWidget(uitools.hdr_widget('default value'),0,4,1,1)
        for ii in range(len(self.op.inputs)):
            rowii = ii+1
            # name
            name = self.op.inputs.keys()[ii] 
            name_widget = uitools.name_widget(name)
            inputs_lo.addWidget( name_widget,rowii,0,1,1 )
            # = 
            eq_widget = uitools.smalltext_widget('=')
            inputs_lo.addWidget( eq_widget,rowii,1,1,1 )
            # src
            src_widget = uitools.src_selection_widget()
            if self.op.input_src[name] is not None:
                src = self.op.input_locator[name].src
            else:
                src = op.no_input 
            src_widget.setCurrentIndex(src)
            inputs_lo.addWidget( src_widget,rowii,2,1,1 )
            # type
            type_widget = uitools.type_selection_widget(src)
            if self.op.input_type[name] is not None:
                tp = self.op.input_locator[name].tp
            else:
                tp = op.none_type
            type_widget.setCurrentIndex(tp)
            inputs_lo.addWidget( type_widget,rowii,3,1,1 )
            # value
            val_widget = QtGui.QLineEdit()
            if self.op.inputs[name] is not None:
                valtext = str(self.op.inputs[name])
            else:
                valtext = 'None'
            inputs_lo.addWidget( val_widget,rowii,4,1,1 )
            self.name_widgets[name] = name_widget 
            self.src_widgets[name] = src_widget
            self.type_widgets[name] = type_widget
            self.val_widgets[name] = val_widget
        inputs_box.setLayout(inputs_lo)
        # set widgets for operation outputs
        outputs_box = QtGui.QGroupBox('OUTPUTS',self.ui)
        outputs_lo = QtGui.QGridLayout(outputs_box)
        if len(self.op.outputs) > 0:
            outputs_lo.addWidget(uitools.r_hdr_widget('name'),0,0,1,1)
            outputs_lo.addWidget(uitools.hdr_widget('description'),0,2,1,1)
        for ii in range(len(self.op.outputs)):
            rowii = ii+1
            # name
            name = self.op.outputs.keys()[ii] 
            name_widget = uitools.name_widget(name)
            outputs_lo.addWidget( name_widget,rowii,0,1,1 )
            # = 
            eq_widget = uitools.smalltext_widget('=')
            outputs_lo.addWidget( eq_widget,rowii,1,1,1 )
            # description 
            desc_widget = uitools.bigtext_widget(self.op.output_doc[name])
            outputs_lo.addWidget( desc_widget,rowii,2,1,1 )
            self.outname_widgets[name] = name_widget 
            self.outdesc_widgets[name] = desc_widget 
        outputs_box.setLayout(outputs_lo)
        # set a widget for the run() function
        #run_widget = QtGui.QTextEdit()
        run_box = QtGui.QGroupBox('RUN METHOD')
        run_lo = QtGui.QGridLayout(run_box)
        self.run_widget = uitools.QSourceEdit()
        run_text = inspect.getsource(self.op.run)
        self.run_widget.setCurrentFont(src_font)
        self.run_widget.setText(run_text)
        self.run_widget.setTabStopWidth(40)
        self.run_widget.setWordWrapMode(QtGui.QTextOption.NoWrap)
        self.run_widget.setSizePolicy(QtGui.QSizePolicy.Preferred,QtGui.QSizePolicy.MinimumExpanding)
        run_lo.addWidget(self.run_widget,0,0,1,1)
        # finally, add the widgets to op_source_layout
        self.ui.op_source_layout.addWidget( self.opname_widget,0,0,1,1 )
        self.ui.op_source_layout.addWidget( path_box,1,0,1,1 )
        self.ui.op_source_layout.addWidget( imports_box,2,0,1,1 )
        self.ui.op_source_layout.addWidget( inputs_box,3,0,1,1 )
        self.ui.op_source_layout.addWidget( outputs_box,4,0,1,1 )
        self.ui.op_source_layout.addWidget( run_box,5,0,1,1 )
        self.set_read_mode()

    def edit_save_op(self):
        if self._edit_mode:
            save_status = self.attempt_save()
            if save_status:
                self.set_read_mode()
        else:
            self.set_edit_mode() 

    def attempt_save(self):
        print('save op')
        ####
        # SAVE
        ####
        saved = True
        return saved

    def set_read_mode(self):
        print('setting to read mode')
        self._edit_mode = False
        #self.ui.op_name.setReadOnly(True)
        for widg in self.import_widgets:
            widg.setReadOnly(True)
        self.ui.edit_save_button.setText("&Edit operation")

    def set_edit_mode(self):
        print('setting to edit mode')
        self._edit_mode = True
        self.ui.edit_save_button.setText("&Save operation")

    def enable_ops(self):
        print('enable ops')

    def setup_ui(self):
        self.ui.setWindowTitle("operation setup")
        self.ui.op_selector.setModel(self.qopman)
        self.ui.op_selector.hideColumn(1)
        self.ui.op_selector.hideColumn(2)
        self.ui.op_selector.clicked.connect( partial(self.get_op) )
        self.ui.edit_save_button.setText("&Edit operation")
        self.ui.enable_op_button.setText("&Enable operations")
        self.ui.edit_save_button.clicked.connect(self.edit_save_op)
        self.ui.enable_op_button.clicked.connect(self.enable_ops)
        self.ui.op_selector.clicked.connect( partial(uitools.toggle_expand,self.ui.op_selector) ) 
        #self.ui.op_name.setAlignment(QtCore.Qt.AlignCenter)
        #self.ui.op_name.setText('-select an operation to view source-')
        self.ui.splitter.setStretchFactor(1,1000)    
        self.ui.setStyleSheet( "QLineEdit { border: none }" + self.ui.styleSheet() )




import os
import string 
from collections import OrderedDict

class DictTree(object):
    """
    A tree as an ordered dictionary (root), 
    extended by embedding other objects 
    that are amenable to tree storage.
    Fetches items by a uri string that is a sequence 
    of dict keys, connected by '.'s.

    Child items (end nodes of the tree)
    can be anything.
    Parent items, in order to index their children,
    must be either lists, dicts, or objects implementing
    keys(), __getitem__(key) and __setitem__(key,value).
    """

    def __init__(self,data={}):
        super(DictTree,self).__init__()
        self._root = OrderedDict()
        if isinstance(data,dict):
            self._root = OrderedDict(data)
        self.bad_chars = string.punctuation 
        self.bad_chars = self.bad_chars.replace('_','')
        self.bad_chars = self.bad_chars.replace('-','')
        self.bad_chars = self.bad_chars.replace('.','')
        self.space_chars = [' ','\t','\n',os.linesep]
        self._all_uris = []

    def __getitem__(self,uri):
        return self.get_from_uri(uri)

    def __setitem__(self,uri,val):
        self.set_uri(uri,val)

    def root_keys(self):
        return self._root.keys()

    #def __len__(self):
    #    return self.n_items()
    #
    #def n_items(self,root_uri=''):
    #    """
    #    Get the total number of data items in the tree.
    #    Only nodes containing data (i.e. end nodes) are counted.
    #    Nodes referencing containers, for example, are not counted. 
    #    """
    #    if root_uri:
    #        itm = self.get_from_uri(root_uri)
    #        prefix = root_uri + '.'
    #    else:
    #        itm = self._root
    #        prefix = ''
    #    if isinstance(itm,list):
    #        return sum([self.n_items(prefix+str(i)) for i in range(len(itm))])
    #    elif isinstance(itm,dict):
    #        return sum([self.n_items(prefix+k) for k in itm.keys()])
    #    else:
    #        # BUG: lacks a solution for simultaneously serving 
    #        # non-dict non-list and non-parent items.
    #        # terminal node: return 1
    #        return 1 

    def delete_uri(self,uri=''):
        """
        Delete the given uri, i.e., 
        remove the corresponding key from the embedded dict.
        This should not be relied on to be fast.
        It has to go through all of the uris to remove children.
        """
        try:
            itm = self._root
            if '.' in uri:
                itm = self.get_from_uri(uri[:uri.rfind('.')])
            path = uri.split('.')
            k = path[-1]
            if k:
                # Note- parent items must implement __getitem__
                del itm[k]
                #itm.pop(k)
                for i,testuri in zip(range(len(self._all_uris))[::-1],self._all_uris[::-1]):
                    if uri in testuri:
                        self._all_uris.pop(i)
        except Exception as ex:
            msg = str('\n[{}] Encountered an error while trying to delete uri {}: \n'
            .format(__name__,uri))
            ex.message = msg + ex.message
            raise ex

    def set_uri(self,uri='',val=None):
        """
        Set the data at the given uri to provided value val.
        """
        try:
            itm = self._root
            if '.' in uri:
                parent_uri = uri[:uri.rfind('.')]
                #if parent_uri in self._all_uris:
                itm = self.get_from_uri(parent_uri)
                #else:
                #    itm = self.build_to_uri(parent_uri)
            k = uri.split('.')[-1]
            # TODO: Is there a more graceful way to handle lists?
            if k:
                if isinstance(itm,list):
                    itm[int(k)] = val
                else:
                    # Note- parent items must implement __setitem__
                    itm[k] = val
                if not uri in self._all_uris:
                    self._all_uris.append(uri)
        except Exception as ex:
            #import pdb; pdb.set_trace()
            msg = str('\n[{}] Encountered an error while trying to set uri {} to val {}: \n'
            .format(__name__,uri,val)) + ex.message
            raise KeyError(msg)

    def get_from_uri(self,uri=''):
        """
        Return the data stored at uri.
        Each data item in the lineage of the uri
        must implement __getitem__() with support for
        string-like keys, unless it is a list,
        in which case the key is cast as int(key)
        before using it as an index in the list.
        """
        try:
            path = uri.split('.')
            itm = self._root 
            #for k in path[:-1]:
            while any(path):
                k = path.pop(0)
                # TODO: Is there a more graceful way to handle lists?
                if isinstance(itm,list):
                    itm = itm[int(k)]
                else:
                    # Note- parent items must implement __getitem__ and keys()
                    if k in itm.keys():
                        itm = itm[k]
                    else:
                        # this could be a dict with a key containing a '.'
                        found = False
                        while not found:
                            k = k+'.'+path.pop(0)                        
                            if k in itm.keys():
                                itm = itm[k]
                                found = True 
                            elif k+'.' in itm.keys():
                                itm = itm[k+'.']
                                found = True           
            return itm
            #k = uri.split('.')[-1]
            #if k == '':
            #    # uri ended with a '.' 
            #    return itm 
            #elif k is not None:
            #    if isinstance(itm,list):
            #        return itm[int(k)]
            #    else:
            #        # Note- parent items must implement __getitem__
            #        return itm[k]
        except Exception as ex:
            #import pdb; pdb.set_trace()
            msg = str('[{}] Encountered an error while fetching uri {}: \n'
            .format(__name__,uri) + ex.message)
            raise KeyError(msg) 

    #def build_to_uri(self,uri=''):
    #    """ 
    #    If the tree does not contain the input uri,
    #    Fill the tree out with dicts
    #    until an empty dict exists at the given uri. 
    #    """
    #    try:
    #        itm = self._root 
    #        if '.' in uri:
    #            parent_uri = uri[:uri.rfind('.')]
    #            if parent_uri in self._all_uris:
    #                itm = self.get_from_uri(parent_uri)
    #            else:
    #                itm = self.build_to_uri(parent_uri)
    #        k = uri.split('.')[-1]
    #        if k == '':
    #            return itm 
    #        elif k is not None:
    #            # Note- parent items must implement __setitem__
    #            if not uri in self._all_uris:
    #                self._all_uris.append(uri)
    #            itm[k] = OrderedDict()
    #            return itm[k]
    #    except Exception as ex:
    #        msg = str('[{}] Encountered an error while trying to build uri {}: \n'
    #        .format(__name__,uri) + ex.message)
    #        raise KeyError(msg) 
        
    #def list_child_tags(self,parent_uri=''):
    #    if parent_uri:
    #        p_itm = self.get_from_uri(parent_uri)
    #    else:
    #        p_itm = self._root
    #    try:
    #        # TODO: Is there a more graceful way to handle lists?
    #        if isinstance(p_itm,list):
    #            return [str(i) for i in range(len(p_itm))]
    #        elif isinstance(p_itm,dict):
    #            return p_itm.keys()
    #        # BUG: lacks a solution for simultaneously serving 
    #        # non-dict non-list and non-parent items.
    #    except Exception as ex:
    #        msg = str('[{}] Encountered an error while trying to get child tags from parent uri {}: \n'
    #        .format(__name__,parent_uri) + ex.message)
    #        raise KeyError(msg) 

    #def list_uris(self,root_uri=''):
    #    if root_uri:
    #        itm = self.get_from_uri(root_uri)
    #        l = [root_uri]
    #        prefix = root_uri+'.'
    #    else:
    #        itm = self._root
    #        l = []
    #        prefix = ''
    #    # TODO: Is there a more graceful way to handle lists?
    #    if isinstance(itm,list):
    #        for i in range(len(itm)):
    #            l = l + self.list_uris(prefix+str(i))
    #    elif isinstance(itm,dict):
    #        for k in itm.keys():
    #            l = l + self.list_uris(prefix+k)
    #        # BUG: lacks a solution for simultaneously serving 
    #        # non-dict non-list and non-parent items.
    #    return l
            
    def is_uri_valid(self,uri):
        """
        Check for validity of a uri. 
        Uris may contain upper case letters, lower case letters, 
        numbers, dashes (-), and underscores (_). 
        Periods (.) are used as delimiters between tags in the uri.
        Any whitespace or any character in the string.punctuation library
        (other than -, _, or .) results in an invalid uri.
        """
        #if parent is None:
        #    parent = self.root_index()
        if (any(map(lambda s: s in uri,self.space_chars))
            or any(map(lambda s: s in uri,self.bad_chars))):
            return False 
        return True 

    def is_tag_valid(self,tag):
        """
        Check for validity of a tag.
        The conditions for a valid tag are the same as for a valid uri,
        except that a tag should not contain period (.) characters.
        """
        if '.' in tag:
            return False 
        else:
            return self.is_uri_valid(tag)

    def is_uri_unique(self,uri):
        """
        Check for uniqueness of a uri. 
        """
        return uri not in self._all_uris
        #if parent is None:
        #    parent = self.root_index()
        #if uri in self._all_uris:
        #    return False 
        #else:
        #    return True 

    def uri_error(self,uri):
        """Provide a human-readable error message for bad uris."""
        if not uri:
            err_msg = 'uri is blank.'
        elif any(map(lambda s: s in uri,self.space_chars)):
            err_msg = 'uri contains whitespace.'
        elif any(map(lambda s: s in uri,self.bad_chars)):
            err_msg = 'uri contains special characters.'
        else:
            err_msg = 'Unforeseen uri error.'
        return str('uri error for {}: \n{}\n'.format(uri,err_msg))

    def tag_error(self,tag):
        """Provide a human-readable error message for bad tags."""
        if '.' in tag:
            return 'tag error for {}: \ntag contains a period (.)\n'.format(tag)
        else:
            return self.uri_error(tag)

    def contains_uri(self,uri):
        """Returns whether or not input uri points to an item in this tree."""
        return uri in self._all_uris
        #if not uri:
        #    return False
        #path = uri.split('.')
        #p_idx = QtCore.QModelIndex()
        #for itemuri in path:
        #    try:
        #        row = self.list_child_tags(p_idx).index(itemuri)
        #    except ValueError as ex:
        #        return False
        #    idx = self.index(row,0,p_idx)
        #    p_idx = idx
        #return True

    def make_unique_uri(self,prefix):
        """
        Generate the next unique uri from prefix by appending '_x' to it, 
        where x is a minimal nonnegative integer.
        """
        suffix = 0
        gooduri = False
        urilist = self._all_uris
        while not gooduri:
            testuri = prefix+'_{}'.format(suffix)
            if not testuri in urilist: 
                gooduri = True
            else:
                suffix += 1
        return testuri 

    def print_tree(self,root_uri='',rowprefix=''):
        """
        Print the content of the tree rooted at root_uri,
        with each row of the string preceded by rowprefix.
        """
        if root_uri:
            itm = self.get_from_uri(root_uri)
        else:
            itm = self._root
        if isinstance(itm,dict):
            tree_string = '\n'
            for k,x in itm.items():
                x_tree = self.print_tree(root_uri+'.'+k,rowprefix+'\t')
                tree_string = tree_string+rowprefix+'{}: {}\n'.format(k,x_tree)
        elif isinstance(itm,list):
            tree_string = '\n'
            for i,x in zip(range(len(itm)),itm):
                x_tree = self.print_tree(root_uri+'.'+str(i),rowprefix+'\t')
                tree_string = tree_string+rowprefix+'{}: {}\n'.format(i,x_tree)
        else:
            return '{}'.format(itm)
        return tree_string

    def print_tree(self,root_uri='',rowprefix=''):
        """
        Print the content of the tree rooted at root_uri,
        with each row of the string preceded by rowprefix.
        """
        if root_uri:
            itm = self.get_from_uri(root_uri)
        else:
            itm = self._root
        if isinstance(itm,dict):
            tree_string = '\n'
            for k,x in itm.items():
                x_tree = self.print_tree(root_uri+'.'+k,rowprefix+'\t')
                tree_string = tree_string+rowprefix+'{}: {}\n'.format(k,x_tree)
        elif isinstance(itm,list):
            tree_string = '\n'
            for i,x in zip(range(len(itm)),itm):
                x_tree = self.print_tree(root_uri+'.'+str(i),rowprefix+'\t')
                tree_string = tree_string+rowprefix+'{}: {}\n'.format(i,x_tree)
        else:
            return '{}'.format(itm)
        return tree_string

#        if parent.isValid():
#            itm = self.get_item(parent)
#            tree_string = tree_string+rowprefix+str(itm.data)+'\n'
#            for j in range(itm.n_children()):
#                tree_string = tree_string + self.print_tree(rowprefix+'\t',self.index(j,0,parent))
#                l.append(root_uri+'.'+str(i))
#                l = l + self.list_uris(root_uri+'.'+str(i))
#
            



from collections import OrderedDict
import glob

from ... import Operation as op
from ...Operation import Batch
from ... import optools

class BatchPostProcess(Batch):
    """
    Take the batch output (list of dicts) from a previously completed Batch,
    and use each dict to form inputs for the execution of a post-processing workflow.
    For each input to be taken from the dict, two uris are needed:
    one to locate it within the (previous) batch outputs, 
    and another to specify where it will be fed to the (current) workflow.
    Collect specified outputs from the workflow for each of the inputs.
    """

    def __init__(self):
        input_names = ['batch_output','batch_uris','input_routes','batch_ops','saved_items']
        output_names = ['batch_inputs','batch_outputs']
        super(BatchPostProcess,self).__init__(input_names,output_names)
        self.input_doc['batch_output'] = ''
        self.input_doc['batch_uris'] = str('uris from batch_output at which the inputs for the current workflow should be found- '
        + ' one batch_uri is expected for each of the input_routes')
        self.input_doc['input_routes'] = 'inputs found from batch_uris are directed to these inputs, which should all be within in batch_ops'
        self.input_doc['batch_ops'] = str('list of workflow uris pointing to Operations to be included in batch execution- '
        + 'the order of entries is unimportant, as the proper execution stack is resolved at runtime')
        self.input_doc['saved_items'] = 'list of uris to be saved in the batch_outputs'
        self.output_doc['batch_inputs'] = 'list of dicts of [input_route:input_value] as determined by batch_uris and input_routes'
        self.output_doc['batch_outputs'] = 'list of dicts of [output_route:output_value] for all saved_items '
        self.input_src['batch_output'] = op.plugin_input
        self.input_src['batch_uris'] = op.wf_input 
        self.input_src['input_routes'] = op.wf_input 
        self.input_src['batch_ops'] = op.wf_input 
        self.input_src['saved_items'] = op.wf_input 
        self.input_type['batch_output'] = op.ref_type
        self.input_type['batch_uris'] = op.path_type
        self.input_type['input_routes'] = op.path_type
        self.input_type['batch_ops'] = op.path_type 
        self.input_type['saved_items'] = op.path_type 
        self.inputs['batch_uris'] = []
        self.inputs['input_routes'] = []
        self.inputs['batch_ops'] = []
        self.inputs['saved_items'] = []
        
    def run(self):
        """
        Build a list of [uri:value] dicts to be used in the workflow.
        """
        b_prev = self.inputs['batch_output']
        b_uris = self.inputs['batch_uris']
        inp_rts = self.inputs['input_routes']
        input_dict_list = []
        output_dict_list = []
        for b_out in b_prev:
            inp_dict = OrderedDict() 
            for b_uri,inp_rt in zip(b_uris,inp_rts):
                inp_dict[inp_rt] = optools.get_uri_from_dict(b_uri,b_out)
            input_dict_list.append(inp_dict)
            output_dict_list.append(OrderedDict())
        self.outputs['batch_inputs'] = input_dict_list
        # Instantiate the batch_outputs list
        self.outputs['batch_outputs'] = output_dict_list 

    def input_list(self):
        return self.outputs['batch_inputs']

    def output_list(self):
        return self.outputs['batch_outputs']

    def input_routes(self):
        """Provide the input route- currently batch execution expects a list."""
        if isinstance(self.inputs['input_routes'],list):
            return self.inputs['input_routes']
        else: 
            return [self.inputs['input_routes']]

    def batch_ops(self):
        """Provide a list of uri's of ops to be included in batch execution"""
        if isinstance(self.inputs['batch_ops'],list):
            return self.inputs['batch_ops']
        else: 
            return [self.inputs['batch_ops']]

    def saved_items(self):
        """List uris to be saved/stored after execution"""
        if isinstance(self.inputs['saved_items'],list):
            return self.inputs['saved_items']
        else: 
            return [self.inputs['saved_items']]

    def batch_outputs_tag(self):
        return 'batch_outputs'

    def set_batch_ops(self,wf=None):
        data = optools.locate_input(self.input_locator['batch_ops'],wf)
        self.input_locator['batch_ops'].data = data 
        self.inputs['batch_ops'] = data

    def set_input_routes(self,wf=None):
        data = optools.locate_input(self.input_locator['input_routes'],wf)
        self.input_locator['input_routes'].data = data 
        self.inputs['input_routes'] = data








from ... import Operation as op
from ...Operation import Realtime 
from ... import optools

class RealtimeFromFiles(Realtime):
    """
    Provides inputs to be used in repeated execution of a workflow
    from files with names matching a regex, as they arrive in a specified directory.
    Collects the outputs produced for each of the inputs.
    """

    def __init__(self):
        input_names = ['dir_path','regex','new_files_only','input_route','realtime_ops','saved_items']
        output_names = ['realtime_inputs','realtime_outputs']
        super(RealtimeFromFiles,self).__init__(input_names,output_names)
        self.input_doc['dir_path'] = 'path to directory where files will be written and then used as input'
        self.input_doc['regex'] = 'string with * wildcards used to filter or locate input files'
        self.input_doc['new_files_only'] = 'if true, ignore existing files and only process new arrivals'
        self.input_doc['input_route'] = 'inputs constructed by the realtime executor are directed to this uri'
        self.input_doc['realtime_ops'] = str('list of ops to be included in the realtime execution- '
        + 'the order of operations in realtime_ops is unimportant, as the proper execution stack is resolved at runtime')
        self.input_doc['saved_items'] = 'list of ops to be saved in the realtime_outputs'
        self.output_doc['realtime_inputs'] = str('iterator over dicts of [input_route:input_value] '
        + 'generated in real time from the local filesystem')
        self.output_doc['realtime_outputs'] = 'list of dicts of [output_route:output_value]'
        self.input_src['dir_path'] = op.fs_input
        self.input_src['regex'] = op.text_input 
        self.input_src['new_files_only'] = op.text_input 
        self.input_src['input_route'] = op.wf_input 
        self.input_src['realtime_ops'] = op.wf_input 
        self.input_src['saved_items'] = op.wf_input 
        self.input_type['dir_path'] = op.path_type
        self.input_type['regex'] = op.str_type
        self.input_type['new_files_only'] = op.bool_type
        self.input_type['input_route'] = op.path_type
        self.input_type['realtime_ops'] = op.path_type 
        self.input_type['saved_items'] = op.path_type 
        self.inputs['regex'] = '*.tif' 
        self.inputs['new_files_only'] = True 
        self.inputs['realtime_ops'] = []
        self.inputs['saved_items'] = []
        
    def run(self):
        """
        This should create an iterator 
        whose next() gives a {uri:value} dict 
        built from the latest-arrived file 
        """
        dirpath = self.inputs['dir_path']
        rx = self.inputs['regex']
        inproute = self.inputs['input_route']
        process_existing_files = not self.inputs['new_files_only']
        self.outputs['realtime_inputs'] = optools.FileSystemIterator(dirpath,rx,process_existing_files)
        self.outputs['realtime_outputs'] = [] 

    def input_iter(self):
        return self.outputs['realtime_inputs']

    def output_list(self):
        return self.outputs['realtime_outputs']

    def input_routes(self):
        """Use the Realtime.input_locators to list uri's of all input routes- must return list."""
        if isinstance(self.inputs['input_route'],list):
            return self.inputs['input_route']
        else:
            return [self.inputs['input_route']]

    def batch_ops(self):
        """Use the Realtime.input_locator to list uri's of ops to be saved/stored after execution"""
        if isinstance(self.inputs['realtime_ops'],list):
            return self.inputs['realtime_ops']
        else:
            return [self.inputs['realtime_ops']]

    def saved_items(self):
        """Use the Realtime.input_locator to list uri's of ops to be included in realtime execution"""
        if isinstance(self.inputs['saved_items'],list):
            return self.inputs['saved_items']
        else:
            return [self.inputs['saved_items']]

    def batch_outputs_tag(self):
        return 'realtime_outputs'

    @staticmethod
    def delay():
        """Amount of time to wait between execution attempts, in milliseconds"""
        return 1000

    def set_batch_ops(self,wf=None):
        data = optools.locate_input(self.input_locator['realtime_ops'],wf)
        self.inputs['realtime_ops'] = data
        #self.input_locator['realtime_ops'].data = data 

    def set_input_routes(self,wf=None):
        data = optools.locate_input(self.input_locator['input_route'],wf)
        self.inputs['input_route'] = data
        #self.input_locator['input_route'].data = data 





import numpy as np
from collections import OrderedDict

from ... import Operation as op
from ...Operation import Operation
from ....tools import saxstools

class SpectrumProfiler(Operation):
    """
    This operation profiles a SAXS spectrum (I(q) vs. q)
    to determine some characteristics of the sample.
    Based on some measures of the overall shape of the spectrum,
    the spectrum is tested for scattering from precursors 
    (approximated as small dilute monodisperse spheres),
    scattering from dilute nanoparticles of various form factors 
    (currently only spheres),
    diffraction peaks
    (Voigt-like profiles due to crystalline arrangements),
    or some combination of the three.
    TODO: document algorithm here.

    Output a return code
    and a dictionary of the results.

    This Operation is somewhat robust for noisy data,
    but any preprocessing (background subtraction, smoothing, or other cleaning)
    should be performed beforehand. 
    """

    def __init__(self):
        input_names = ['q', 'I', 'dI']
        output_names = ['features']
        super(SpectrumProfiler, self).__init__(input_names, output_names)
        self.input_doc['q'] = '1d array of wave vector values in 1/Angstrom units'
        self.input_doc['I'] = '1d array of intensity values I(q)'
        self.input_doc['dI'] = 'optional- 1d array of intensity uncertainty values I(q)'
        self.output_doc['features'] = str('dict profiling the input spectrum. '
        + 'Dict keys and descriptions: \n'
        + 'ERROR_MESSAGE: Any errors or warnings are reported here. \n'
        + 'bad_data_flag: Boolean indicating that the spectrum is unfamiliar or mostly made of noise. \n'
        + 'precursor_flag: Boolean indicating presence of precursor terms. \n'
        + 'form_flag: Boolean indicating presence of form factor terms. \n'
        + 'form_id: Form factor identity, e.g. sphere, cube, rod, etc. \n'
        + 'structure_flag: Boolean indicating presence of structure factor terms. \n'
        + 'structure_id: Structure identity, e.g. fcc, hcp, bcc, etc. \n'
        + 'low_q_ratio: fraction of integrated intensity for q<0.4 '
        + 'high_q_ratio: fraction of integrated intensity for q>0.4 '
        + 'low_q_logratio: fraction of total log(I)-min(log(I)) over q<0.4 '
        + 'high_q_logratio: fraction of total log(I)-min(log(I)) over q>0.4 '
        + 'high_freq_ratio: Ratio of the upper half to the lower half '
        + 'of the power spectrum of the discrete fourier transform of the intensity. \n'
        + 'fluctuation_strength: Integrated fluctuation of intensity '
        + '(sum of difference in intensity between adjacent points '
        + 'taken only where this difference changes sign), '
        + 'divided by the range (maximum minus minimum) of intensity. \n'
        + 'Imax_over_Imean: maximum intensity divided by mean intensity. \n'
        + 'Imax_over_Imean_local: maximum intensity divided by '
        + 'mean intensity over q values within 10% of the q value of the maximum. \n'
        + 'Imax_over_Ilowq: Maximum intensity divided by mean intensity over q<0.1. \n'
        + 'Imax_over_Ihighq: Maximum intensity divided by mean intensity over q>0.4. \n'
        + 'Ilowq_over_Ihighq: Mean intensity for q<0.1 divided by mean intensity for q>0.4. \n'
        + 'low_q_logcurv: Curvature of parabola fit to log(I) versus log(q) for q<0.1. \n'
        + 'q_Imax: q value of the maximum intensity. ')
        #+ 'bin_strengths: (not implemented). \n'
        #+ 'q_corr: Pearson correlation between I(q) and q. \n'
        #+ 'qsquared_corr: Pearson correlation between I(q) and q^2. \n'
        #+ 'cos2q_corr: Pearson correlation between I(q) and (cos(q))^2. \n'
        #+ 'cosq_corr: Pearson correlation between I(q) and cos(q). \n'
        #+ 'invq_corr: Pearson correlation between I(q) and 1/q. \n'
        #+ 'q_logcorr: Same as q_corr, but with log(I(q)). \n'
        #+ 'qsquared_logcorr: Same as qsquared_corr, but with log(I(q)). \n'
        #+ 'cos2q_logcorr: Same as cos2q_corr, but with log(I(q)). \n'
        #+ 'cosq_logcorr: Same as cosq_corr, but with log(I(q)). \n'
        #+ 'invq_logcorr: Same as invq_corr, but with log(I(q)).')
        self.input_src['q'] = op.wf_input
        self.input_src['I'] = op.wf_input
        self.input_src['dI'] = op.no_input
        self.input_type['q'] = op.ref_type
        self.input_type['I'] = op.ref_type

    def run(self):
        q, I = self.inputs['q'], self.inputs['I']
        dI = self.inputs['dI']
        # Set return code to 1 (error) by default;
        # if execution finishes, set it to 0

        ### amplitude analysis:
        idxmax = np.argmax(I)
        Imax = I[idxmax] 
        q_Imax = q[idxmax]
        idxmin = np.argmin(I)
        Imin = I[idxmin]
        Irange = Imax - Imin
        Imean = np.mean(I)
        Imax_over_Imean = float(Imax)/float(Imean)
        idx_around_max = ((q > 0.9*q_Imax) & (q < 1.1*q_Imax))
        Imean_around_max = np.mean(I[idx_around_max])
        Imax_over_Imean_local = Imax / Imean_around_max

        ### fourier analysis
        n_q = len(q)
        fftI = np.fft.fft(I)
        ampI = np.abs(fftI)
        powI = np.abs(fftI)**2
        high_freq_ratio = np.sum(powI[n_q/4:n_q/2])/np.sum(powI[1:n_q/4])
        #high_freq_ratio = np.sum(ampI[n_q/4:n_q/2])/np.sum(ampI[1:n_q/4])

        ### fluctuation analysis
        # array of the difference between neighboring points:
        nn_diff = I[1:]-I[:-1]
        # keep indices where the sign of this difference changes.
        # also keep first index
        nn_diff_prod = nn_diff[1:]*nn_diff[:-1]
        idx_keep = np.hstack((np.array([True]),nn_diff_prod<0))
        fluc = np.sum(np.abs(nn_diff[idx_keep]))
        #fluctuation_strength = fluc/Irange
        fluctuation_strength = fluc/Imean

	    # correlations on intensity
        q_corr = saxstools.compute_pearson(I,np.linspace(0,1,n_q))
        qsquared_corr = saxstools.compute_pearson(I,np.linspace(0,1,n_q)**2)
        cosq_corr = saxstools.compute_pearson(I,np.cos(q*np.pi/(2*q[-1])))
        cos2q_corr = saxstools.compute_pearson(I,np.cos(q*np.pi/(2*q[-1]))**2)
        invq_corr = saxstools.compute_pearson(I,q**-1)
        #invq4_corr = saxstools.compute_pearson(I,q**-4)

        # correlations on log intensity
        idx_nz = np.invert(I <= 0)
        q_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , np.linspace(0,1,n_q)[idx_nz] )
        qsquared_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , (np.linspace(0,1,n_q)[idx_nz])**2 )
        cosq_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , np.cos(q*np.pi/(2*q[-1]))[idx_nz] )
        cos2q_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , (np.cos(q*np.pi/(2*q[-1]))[idx_nz])**2 )
        invq_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , q[idx_nz]**-1 )

        ### bin-integrated intensity analysis
        # TODO: standardize these bins in q space
        #qmin, qmax = q[0], q[-1]
        #qrange = qmax-qmin
        #bin_strengths = np.zeros(10)
        #for i in range(10):
        #    qmini, qmaxi = qmin+i*qrange/10,qmin+(i+1)*qrange/10
        #    idxi = list((q>=qmini) & (q<qmaxi))
        #    idxi_shift = [False]+idxi[:-1]
        #    idxi = np.array(idxi,dtype=bool)
        #    idxi_shift = np.array(idxi_shift,dtype=bool)
        #    dqi = q[ idxi_shift ] - q[ idxi ]
        #    Ii = I[ idxi ]
        #    bin_strengths[i] = np.sum(Ii * dqi) / (qmaxi-qmini)
        idx_nz = np.invert((I==0))
        q_nz = q[idx_nz] 
        I_nz_log = np.log(I[idx_nz])
        # make values positive:
        I_nz_log = I_nz_log-np.min(I_nz_log)
        I_logsum = np.sum(I_nz_log)
        low_q_logratio = np.sum(I_nz_log[(q_nz<0.4)])/I_logsum
        high_q_logratio = np.sum(I_nz_log[(q_nz>=0.4)])/I_logsum
        I_sum = np.sum(I)
        low_q_ratio = np.sum(I[(q<0.4)])/I_sum
        high_q_ratio = np.sum(I[(q>=0.4)])/I_sum

        ### curve shape analysis
        lowq_idx = q<0.1
        highq_idx = q>0.4
        lowq = q[lowq_idx]
        highq = q[highq_idx]
        I_lowq = I[lowq_idx]
        I_highq = I[highq_idx]
        I_lowq_mean = np.mean(I_lowq)
        I_highq_mean = np.mean(I_highq)
        lowq_mean = np.mean(lowq)
        lowq_std = np.std(lowq)
        I_lowq_std = np.std(I_lowq)
        I_lowq_s = I_lowq/I_lowq_std
        lowq_s = (lowq - lowq_mean)/lowq_std
        #p_lowq = saxstools.fit_with_slope_constraint(lowq_s,np.log(I_lowq_s),-1*lowq_mean/lowq_std,0,3) 
        #p_lowq = saxstools.fit_with_slope_constraint(lowq_s,np.log(I_lowq_s),lowq_s[-1],0,3) 
        p_lowq = np.polyfit(lowq_s,np.log(I_lowq_s),2)
        low_q_logcurv = p_lowq[0]
        Imax_over_Ilowq = float(Imax)/I_lowq_mean
        Imax_over_Ihighq = float(Imax)/I_highq_mean
        Ilowq_over_Ihighq = I_lowq_mean/I_highq_mean

        # Flagging bad data: 
        # Data with high noise are bad.
        # Data that are totally flat or increasing in q are bad.
        # Data that do not have downward curvature at low q are often bad, but not always.
        bad_data_flag = ( fluctuation_strength > 20 
                        or low_q_ratio/high_q_ratio < 1
                        or Ilowq_over_Ihighq < 10 )
                        #or (low_q_logcurv > 0 and Ilowq_over_Ihighq < 10))

        form_id = None 
        structure_id = None 
        if bad_data_flag:
            form_flag = False
            precursor_flag = False
            structure_flag = False
        else:
            # Flagging form factor:
            # Intensity should be quite decreasing in q.
            # Low-q region should be quite flat.
            # Low-q mean intensity should be much larger than low-q fluctuations.
            form_flag = low_q_ratio/high_q_ratio > 10 
            if form_flag:
                # TODO: determine form factor here
                form_id = 'sphere'
 
            # Flagging precursors: 
            # Intensity should decrease in q, at least mildly.
            # Intensity should decrease more sharply at high q.
            # Low-q region of spectrum should be quite flat.
            # Noise levels may be high if only precursors are present.
            # More high-q intensity than form factor alone.
            precursor_flag = low_q_ratio/high_q_ratio > 2 and high_q_ratio > 1E-3 

            # Flagging structure:
            # Structure is present if max intensity occurs outside the low-q region. 
            # Maximum intensity should be large relative to its 'local' mean intensity. 
            structure_flag = Imax_over_Imean_local > 2 
            if structure_flag:
                # TODO: determine structure factor here
                structure_id = 'fcc'
        d_r = OrderedDict() 
        #d_r['bin_strengths'] = bin_strengths
        #d_r['bin_strengths'] = None 
        d_r['bad_data_flag'] = bad_data_flag
        d_r['precursor_flag'] = precursor_flag
        d_r['form_flag'] = form_flag
        d_r['structure_flag'] = structure_flag
        d_r['structure_id'] = structure_id 
        d_r['form_id'] = form_id 
        d_r['low_q_logcurv'] = low_q_logcurv
        d_r['Imax_over_Imean'] = Imax_over_Imean
        d_r['Imax_over_Imean_local'] = Imax_over_Imean_local
        d_r['Imax_over_Ilowq'] = Imax_over_Ilowq 
        d_r['Imax_over_Ihighq'] = Imax_over_Ihighq 
        d_r['Ilowq_over_Ihighq'] = Ilowq_over_Ihighq 
        d_r['low_q_logratio'] = low_q_logratio 
        d_r['high_q_logratio'] = high_q_logratio 
        d_r['low_q_ratio'] = low_q_ratio 
        d_r['high_q_ratio'] = high_q_ratio 
        d_r['high_freq_ratio'] = high_freq_ratio 
        d_r['fluctuation_strength'] = fluctuation_strength
        d_r['q_Imax'] = q_Imax
        #d_r['q_logcorr'] = q_logcorr
        #d_r['qsquared_logcorr'] = qsquared_logcorr
        #d_r['cos2q_logcorr'] = cos2q_logcorr
        #d_r['cosq_logcorr'] = cosq_logcorr
        #d_r['invq_logcorr'] = invq_logcorr
        #d_r['q_corr'] = q_corr
        #d_r['qsquared_corr'] = qsquared_corr
        #d_r['cos2q_corr'] = cos2q_corr
        #d_r['cosq_corr'] = cosq_corr
        #d_r['invq_corr'] = invq_corr
        self.outputs['features'] = d_r 




import os

from pypif import pif
import pypif.obj as pifobj 

from ... import Operation as op
from ...Operation import Operation
        
class ShipToDataSet(Operation):
    """
    Take a pypif.obj.System object and ship it to a given Citrination data set.    
    """

    def __init__(self):
        input_names = ['pif','client','dsid','json_path','keep_json','ship_flag']
        output_names = ['response']
        super(ShipToDataSet,self).__init__(input_names,output_names)
        self.input_doc['pif'] = 'A pypif.obj.System object or an array/list thereof'
        self.input_doc['client'] = 'A running Citrination client' 
        self.input_doc['dsid'] = 'Data set ID where the pif record(s) will be stored on Citrination' 
        self.input_doc['json_path'] = 'Filesystem path where the json of the pif(s) will be saved' 
        self.input_doc['keep_json'] = 'Flag for whether or not to keep the json file' 
        self.input_doc['ship_flag'] = 'Flag for shipping the pif- set to False for a dry run' 
        self.output_doc['response'] = 'The Citrination server response to the shipment'
        self.input_src['pif'] = op.wf_input
        self.input_src['client'] = op.plugin_input
        self.input_src['dsid'] = op.text_input
        self.input_src['json_path'] = op.fs_input
        self.input_src['keep_json'] = op.text_input
        self.input_src['ship_flag'] = op.text_input
        self.input_type['pif'] = op.ref_type
        self.input_type['client'] = op.ref_type
        self.input_type['dsid'] = op.int_type
        self.input_type['json_path'] = op.path_type
        self.input_type['keep_json'] = op.bool_type
        self.input_type['ship_flag'] = op.bool_type
        self.inputs['ship_flag'] = False
        self.inputs['keep_json'] = False

    def run(self):
        cl = self.inputs['client'] 
        dsid = self.inputs['dsid'] 
        p = self.inputs['pif']        
        json_path = self.inputs['json_path']
        json_flag = self.inputs['keep_json']
        ship_flag = self.inputs['ship_flag']
        try:
            if isinstance(p,pifobj.ChemicalSystem):
                json_file = json_path+'/'+p.uid+'.json'
            else:
                json_file = json_path+'/pif_tmp.json'
            # make p an array of pifs to get a big json that has all records
            pif.dump(p, open(json_file,'w'))
            if ship_flag:
                r = cl.upload_file(json_file,dataset_id = dsid)
            else:
                r = 'dry run: no shipment occurred. pif object: {}'.format(pif.dumps(p))
            if not json_flag:
                os.remove(json_file) 
        except Exception as ex:
            r = 'An error occurred while shipping. Error message: {}'.format(ex.message)
        self.outputs['response'] = r





from __future__ import print_function
from collections import OrderedDict
import copy
import traceback
import time

from .Workflow import Workflow
from ..operations import Operation as op
from ..operations.Operation import Operation, Batch, Realtime        
from ..operations import optools
from .. import pawstools

class WfManager(object):
    """
    Manager for paws Workflows. 
    Stores a list of Workflow objects, 
    performs operations on them.
    Keeps a reference to a PluginManager
    for access to PawsPlugins.
    """

    def __init__(self,plugin_manager):
        #super(WfManager,self).__init__()
        self.workflows = OrderedDict() 
        self.plugman = plugin_manager
        self.logmethod = print 

    #def __getitem__(self,key):
    #    if key in self.workflows.keys():
    #        return self.workflows[key]
    #    else:
    #        raise KeyError('[{}] WfManager does not recognize workflow name {}.'
    #        .format(__name__,key))

    def n_wf(self):
        return len(self.workflows)

    def write_log(self,msg):
        self.logmethod(msg)
        #if self.logmethod:
        #else:
        #    print('- '+pawstools.timestr()+': '+msg)

    def add_wf(self,wfname):
        """
        Add a workflow to self.workflows, with key specified by wfname.
        If wfname is not unique (i.e. a workflow with that name already exists),
        this method will overwrite the existing workflow with a new one.
        """
        wf = Workflow()
        self.workflows[wfname] = wf

    def run_wf(self,wfname):
        """
        Serially execute the operations of WfManager.workflows[wfname].
        Uses optools.execution_stack() to determine execution order.
        """
        stk,diag = optools.execution_stack(self.workflows[wfname])
        for lst in stk:
            batch_flag = isinstance(self.workflows[wfname].get_data_from_uri(lst[0]),Batch)
            rt_flag = isinstance(self.workflows[wfname].get_data_from_uri(lst[0]),Realtime)
            if not any([batch_flag,rt_flag]):
                self.execute_serial(wfname,lst)
            elif batch_flag:
                self.execute_batch(wfname,lst[0],lst[1])
            elif rt_flag:
                self.execute_realtime(wfname,lst[0],lst[1])

    def execute_batch(self,wfname,batch_op_tag,batch_stk):
        batch_op = self.workflows[wfname].get_data_from_uri(batch_op_tag) 
        optools.load_inputs(batch_op,self.workflows[wfname],self.plugman)
        batch_op.run()
        self.workflows[wfname].set_item(batch_op_tag,batch_op)
        n_batch = len(batch_op.input_list())
        for i in range(n_batch):
            input_dict = batch_op.input_list()[i]
            for uri,val in input_dict.items():
                self.workflows[wfname].set_op_input_at_uri(uri,val)
            self.write_log( 'BATCH EXECUTION {} / {}'.format(i+1,n_batch) )
            for batch_lst in batch_stk:
                self.execute_serial(wfname,batch_lst)
            saved_items_dict = OrderedDict()
            for uri in batch_op.saved_items():
                # TODO # BUG: there is the chance for infinite recursion here
                # if the batch is asked to save an upstream item?
                save_data = copy.deepcopy(self.workflows[wfname].get_data_from_uri(uri))
                save_dict = self.uri_to_embedded_dict(uri,save_data) 
                saved_items_dict = self.update_embedded_dict(saved_items_dict,save_dict)
            batch_op.output_list()[i] = saved_items_dict
            # TODO: set a more specific item here to save some tree update time?
            #self.workflows[wfname].set_item(batch_op_tag,batch_op)
            outputs_uri = batch_op_tag+'.'+op.outputs_tag+'.'+batch_op.batch_outputs_tag()+'.'+str(i)
            self.workflows[wfname].set_item(outputs_uri,saved_items_dict)

    def execute_realtime(self,wfname,rt_op_tag,rt_stk):
        rt_op = self.workflows[wfname].get_data_from_uri(rt_op_tag) 
        optools.load_inputs(rt_op,self.workflows[wfname],self.plugman)
        rt_op.run()
        self.workflows[wfname].set_item(rt_op_tag,rt_op)
        keep_running = True
        n_exec = 0
        wait_iter = 0
        while keep_running:
        ##TODO: ways to control the loop exit condition
            vals = rt_op.input_iter().next()
            if not None in vals:
                n_exec += 1
                wait_iter = 0
                inp_dict = OrderedDict( zip(rt_op.input_routes(), vals) )
                for uri,val in inp_dict.items():
                    self.workflows[wfname].set_op_input_at_uri(uri,val) 
                self.write_log( 'REALTIME EXECUTION {}'.format(n_exec))
                for rt_lst in rt_stk:
                    self.execute_serial(wfname,rt_lst)
                saved_items_dict = OrderedDict()
                for uri in rt_op.saved_items():
                    save_data = copy.deepcopy(self.workflows[wfname].get_data_from_uri(uri))
                    save_dict = self.uri_to_embedded_dict(uri,save_data) 
                    saved_items_dict = self.update_embedded_dict(saved_items_dict,save_dict)
                rt_op.output_list().append(saved_items_dict)
                output_uri = rt_op_tag+'.'+op.outputs_tag+'.'+rt_op.batch_outputs_tag()+'.'+str(i)
                self.workflows[wfname].set_item(output_uri,saved_items_dict)
            else:
                if wait_iter == 0:
                    self.write_log( 'Waiting for new inputs...' )
                wait_iter += 1 
                time.sleep(float(rt_op.delay())/1000.0)
            if wait_iter > 1000:
                self.write_log('Waited too long. Exiting...')
                keep_running = False

    def execute_serial(self,wfname,op_list):
        self.write_log('workflow {} running {}'.format(wfname,op_list))
        for op_tag in op_list: 
            op = self.workflows[wfname].get_data_from_uri(op_tag) 
            optools.load_inputs(op,self.workflows[wfname],self.plugman)
            try:
                op.run() 
            except Exception as ex:
                tb = traceback.format_exc()
                self.write_log(str('Operation {} of workflow {} threw an error. '
                + '\nMessage: {} \nTrace: {}').format(op_tag,wfname,ex.message,tb)) 
            self.workflows[wfname].set_item(op_tag,op)

    def uri_to_embedded_dict(self,uri,data=None):
        path = uri.split('.')
        endtag = path[-1]
        d = OrderedDict()
        d[endtag] = data
        for tag in path[:-1][::-1]:
            parent_d = OrderedDict()
            parent_d[tag] = d
            d = parent_d
        return d

    def update_embedded_dict(self,d,d_new):
        for k,v in d_new.items():
            if k in d.keys():
                if isinstance(d[k],dict) and isinstance(d_new[k],dict):
                    # embedded dicts: recurse
                    d[k] = self.update_embedded_dict(d[k],d_new[k])
                else:
                    # existing d[k] is not dict, and d_new[k] is dict: replace  
                    d[k] = d_new[k]
            else:
                # d[k] does not exist: insert
                d[k] = v
        return d

    def load_from_dict(self,wfname,opman,opdict):
        """
        Create a workflow with name wfname.
        If wfname is not unique, self.workflows[wfname] is overwritten.
        Input opdict specifies operation setup,
        where each item in opdict provides enough information
        to get an Operation from OpManager opman
        and set up its Operation.input_locators.
        """
        self.add_wf(wfname)
        for opname, op_setup in opdict.items():
            op = self.build_op_from_dict(op_setup,opman)
            if isinstance(op,Operation):
                self.workflows[wfname].set_item(opname,op)
            else:
                self.write_log('[{}] Failed to load {}.'.format(uri))

    def op_setup_dict(self,op):
        op_mod = op.__module__[op.__module__.find('operations'):]
        op_mod = op_mod[op_mod.find('.')+1:]
        dct = OrderedDict() 
        dct['op_module'] = op_mod
        inp_dct = OrderedDict() 
        for name in op.inputs.keys():
            il = op.input_locator[name]
            inp_dct[name] = {'src':copy.copy(il.src),'tp':copy.copy(il.tp),'val':copy.copy(il.val)}
        dct[op.inputs_tag] = inp_dct 
        return dct

    def build_op_from_dict(self,op_setup,opman):
        op_uri = op_setup['op_module']
        op = opman.get_data_from_uri(op_uri)
        if issubclass(op,Operation):
            op = op()
            op.load_defaults()
            il_setup_dict = op_setup[op.inputs_tag]
            for name in op.inputs.keys():
                if name in il_setup_dict.keys():
                    src = il_setup_dict[name]['src']
                    #if 'tp' in il_setup_dict[name].keys():
                    tp = il_setup_dict[name]['tp']
                    val = il_setup_dict[name]['val']
                    if tp in op.invalid_types[src]:
                        il = op.InputLocator(src,op.none_type,None)
                    else:
                        il = op.InputLocator(src,tp,val)
                    op.input_locator[name] = il
                    # dereference any existing inputs
                    # LAP: commented out bc it should be handled in op.load_defaults()
                    #op.inputs[name] = None
            return op
        else:
            return None

    #def auto_name(self,wfname):
    #    """
    #    Generate the next unique workflow name by appending '_x',
    #    where x is a minimal nonnegative integer.
    #    """
    #    goodname = False
    #    prefix = wfname
    #    idx = 1
    #    while not goodname:
    #        if not wfname in self.workflows.keys():
    #            goodname = True
    #        else:
    #            wfname = prefix+'_{}'.format(idx)
    #            idx += 1
    #    return wfname 





import numpy as np

from .. import Operation as op
from ..Operation import Operation

class WindowZip(Operation):
    """
    From input sequences for x and y, 
    produce an n-by-2 array 
    where x is bounded by the specified limits 
    """

    def __init__(self):
        input_names = ['x','y','x_min','x_max']
        output_names = ['x_window','y_window','x_y_window']
        super(WindowZip,self).__init__(input_names,output_names)        
        self.input_src['x'] = op.wf_input
        self.input_src['y'] = op.wf_input
        self.input_src['x_min'] = op.text_input
        self.input_src['x_max'] = op.text_input
        self.input_type['x'] = op.ref_type
        self.input_type['y'] = op.ref_type
        self.input_type['x_min'] = op.float_type
        self.input_type['x_max'] = op.float_type
        self.inputs['x_min'] = 0.02 
        self.inputs['x_max'] = 0.6 
        self.input_doc['x'] = 'list (or iterable) of x values'
        self.input_doc['y'] = 'list (or iterable) of y values'
        self.input_doc['x_min'] = 'inclusive minimum x value of output'
        self.input_doc['x_max'] = 'inclusive maximum x value of output'
        self.output_doc['x_window'] = 'n-by-1 array of x_min <= x <= x_max'
        self.output_doc['y_window'] = 'n-by-1 array of y for x_min <= x <= x_max'
        self.output_doc['x_y_window'] = 'n-by-2 array with x, y pairs for x_min <= x <= x_max'

    def run(self):
        xvals = self.inputs['x']
        yvals = self.inputs['y']
        x_min = self.inputs['x_min']
        x_max = self.inputs['x_max']
        x_y_window = window_zip(xvals, yvals, x_min, x_max)
        self.outputs['x_window'] = x_y_window[:,0]
        self.outputs['y_window'] = x_y_window[:,1]
        self.outputs['x_y_window'] = x_y_window

def window_zip(x, y, x_min, x_max):
    good = ((x >= x_min) & (x <= x_max))
    return zip(x[good],y[good])

def zip(x, y):
    x_y = np.zeros((x.size, 2))
    x_y[:, 0] = x
    x_y[:, 1] = y
    return x_y

def logsafe_zip(x, y):
    bad = (x <= 0) | (y <= 0) | np.isnan(y)
    return zip(x[~bad], y[~bad])
